from textwrap import dedent

from neuro_flow.commands import CmdProcessor


async def test_empty() -> None:
    async with CmdProcessor() as proc:
        pass

    assert proc.outputs == {}
    assert proc.states == {}


async def test_no_commands() -> None:
    out = []
    async with CmdProcessor() as proc:
        async for line in proc.feed_chunk(b"123\n34"):
            out.append(line)
        async for line in proc.feed_chunk(b"56\n78"):
            out.append(line)
        async for line in proc.feed_eof():
            out.append(line)

    assert proc.outputs == {}
    assert proc.states == {}
    assert b"".join(out).decode("utf-8") == dedent(
        """\
        123
        3456
        78"""
    )


async def test_set_output1() -> None:
    out = []
    async with CmdProcessor() as proc:
        async for line in proc.feed_chunk(b"::set-output name=key1::value1\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"Just a line\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"::set-output name=key2::value2\n"):
            out.append(line)
        async for line in proc.feed_eof():
            out.append(line)

    assert proc.outputs == {"key1": "value1", "key2": "value2"}
    assert proc.states == {}
    assert b"".join(out).decode("utf-8") == "Just a line\n"


async def test_set_output2() -> None:
    out = []
    async with CmdProcessor() as proc:
        async for line in proc.feed_chunk(b"::set-"):
            out.append(line)
        async for line in proc.feed_chunk(b"output name=key1::value1\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"::set-output name=key2::value2\n"):
            out.append(line)
        async for line in proc.feed_eof():
            out.append(line)

    assert proc.outputs == {"key1": "value1", "key2": "value2"}
    assert proc.states == {}
    assert b"".join(out).decode("utf-8") == dedent("")


async def test_set_output3() -> None:
    out = []
    async with CmdProcessor() as proc:
        async for line in proc.feed_chunk(b"::set-"):
            out.append(line)
        async for line in proc.feed_chunk(b"output name=key1::value1\n"):
            out.append(line)
        async for line in proc.feed_chunk(
            b"::set-output name=key2::value2\nJust a line"
        ):
            out.append(line)
        async for line in proc.feed_eof():
            out.append(line)

    assert proc.outputs == {"key1": "value1", "key2": "value2"}
    assert proc.states == {}
    assert b"".join(out).decode("utf-8") == "Just a line"


async def test_save_state() -> None:
    out = []
    async with CmdProcessor() as proc:
        async for line in proc.feed_chunk(b"::save-state name=key1::value1\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"Just a line\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"::save-state name=key2::value2\n"):
            out.append(line)
        async for line in proc.feed_eof():
            out.append(line)

    assert proc.outputs == {}
    assert proc.states == {"key1": "value1", "key2": "value2"}
    assert b"".join(out).decode("utf-8") == "Just a line\n"


async def test_stop_commands() -> None:
    out = []
    async with CmdProcessor() as proc:
        async for line in proc.feed_chunk(b"::stop-commands::resume\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"::save-state name=key1::value1\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"Just a line\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"::resume::\n"):
            out.append(line)
        async for line in proc.feed_chunk(b"::save-state name=key2::value2\n"):
            out.append(line)
        async for line in proc.feed_eof():
            out.append(line)

    assert proc.outputs == {}
    assert proc.states == {"key2": "value2"}
    assert b"".join(out).decode("utf-8") == dedent(
        """\
        ::save-state name=key1::value1
        Just a line
    """
    )



import pathlib
import pytest
import sys
import textwrap
from neuro_sdk import Client
from typing import AsyncContextManager, AsyncIterator, Callable, Mapping

from neuro_flow.batch_runner import (
    ImageRefNotUniqueError,
    build_graphs,
    check_image_refs_unique,
    check_local_deps,
    check_no_cycles,
    iter_flows,
    upload_image_data,
)
from neuro_flow.colored_topo_sorter import CycleError
from neuro_flow.config_loader import BatchLocalCL, ConfigLoader
from neuro_flow.context import EarlyBatchAction, RunningBatchFlow
from neuro_flow.expr import MultiError
from neuro_flow.parser import ConfigDir
from neuro_flow.storage.base import BakeImage, Storage
from neuro_flow.storage.in_memory import InMemoryStorage


if sys.version_info >= (3, 7):  # pragma: no cover
    from contextlib import asynccontextmanager
else:
    from async_generator import asynccontextmanager

BatchClFactory = Callable[[str], AsyncContextManager[ConfigLoader]]


@pytest.fixture
async def batch_cl_factory(
    loop: None,
    assets: pathlib.Path,
    client: Client,
) -> Callable[[str], AsyncContextManager[ConfigLoader]]:
    @asynccontextmanager
    async def _factory(subpath: str = "") -> AsyncIterator[ConfigLoader]:
        config_dir = ConfigDir(
            workspace=assets / subpath,
            config_dir=assets / subpath,
        )
        cl = BatchLocalCL(config_dir, client)
        yield cl
        await cl.close()

    return _factory


@pytest.fixture()
def batch_storage(loop: None) -> Storage:
    return InMemoryStorage()


async def test_iter_flows(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("early_graph") as cl:
        flow = await RunningBatchFlow.create(cl, "batch", "bake-id")
        prefix2flow = {prefix: flow async for prefix, flow in iter_flows(flow)}

        assert prefix2flow[()] == flow
        action_flow = prefix2flow[("second",)]
        assert isinstance(action_flow, EarlyBatchAction)
        assert action_flow._action == (await flow.get_action_early("second"))._action

        action_flow = prefix2flow[("third",)]
        assert isinstance(action_flow, EarlyBatchAction)
        assert action_flow._action == (await flow.get_action_early("third"))._action


async def test_check_no_cycles(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("") as cl:
        flow = await RunningBatchFlow.create(cl, "batch-cycle", "bake-id")
        with pytest.raises(CycleError):
            await check_no_cycles(flow)


async def test_check_cycles(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("") as cl:
        flow = await RunningBatchFlow.create(cl, "batch-cycle", "bake-id")
        with pytest.raises(CycleError):
            await check_no_cycles(flow)


async def test_local_deps_on_remote_1(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("local_actions") as cl:
        flow = await RunningBatchFlow.create(cl, "bad-order", "bake-id")
        with pytest.raises(
            Exception, match=r"Local action 'local' depends on remote task 'remote'"
        ):
            await check_local_deps(flow)


async def test_local_deps_on_remote_2(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("local_actions") as cl:
        flow = await RunningBatchFlow.create(cl, "bad-order-through-action", "bake-id")
        with pytest.raises(
            Exception,
            match=r"Local action 'local' depends on "
            r"remote task 'call_action.remote_task'",
        ):
            await check_local_deps(flow)


async def test_graphs(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("") as cl:
        flow = await RunningBatchFlow.create(cl, "batch-action-call", "bake-id")
        graphs = await build_graphs(flow)
        assert graphs == {
            (): {("test",): set()},
            ("test",): {
                ("test", "task_1"): set(),
                ("test", "task_2"): {("test", "task_1")},
            },
        }


async def test_early_graph(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("early_graph") as cl:
        flow = await RunningBatchFlow.create(cl, "batch", "bake-id")
        graphs = await build_graphs(flow)
        assert graphs == {
            (): {
                ("first_ac",): set(),
                ("second",): {("first_ac",)},
                ("third",): {("first_ac",)},
            },
            ("first_ac",): {("first_ac", "task_2"): set()},
            ("second",): {
                ("second", "task-1-o3-t3"): set(),
                ("second", "task-1-o1-t1"): set(),
                ("second", "task-1-o2-t1"): set(),
                ("second", "task-1-o2-t2"): set(),
                ("second", "task_2"): {
                    ("second", "task-1-o3-t3"),
                    ("second", "task-1-o1-t1"),
                    ("second", "task-1-o2-t1"),
                    ("second", "task-1-o2-t2"),
                },
            },
            ("third",): {
                ("third", "task-1-o3-t3"): set(),
                ("third", "task-1-o1-t1"): set(),
                ("third", "task-1-o2-t1"): set(),
                ("third", "task-1-o2-t2"): set(),
                ("third", "task_2"): {
                    ("third", "task-1-o3-t3"),
                    ("third", "task-1-o1-t1"),
                    ("third", "task-1-o2-t1"),
                    ("third", "task-1-o2-t2"),
                },
            },
        }


async def test_check_image_refs_unique(batch_cl_factory: BatchClFactory) -> None:
    async with batch_cl_factory("batch_images") as cl:
        flow = await RunningBatchFlow.create(cl, "duplicate_ref", "bake-id")
        with pytest.raises(MultiError) as err_info:
            await check_image_refs_unique(flow)
    err = err_info.value
    assert isinstance(err, MultiError)
    assert len(err.errors) == 1
    inner_err = err.errors[0]

    assert isinstance(inner_err, ImageRefNotUniqueError)
    ws = cl.workspace
    assert str(inner_err) == textwrap.dedent(
        f"""\
        Image with ref 'image:banana1' defined multiple times with different attributes:
        at "{ws / "duplicate_ref.yml"}", line 4, column 5 with params:
          context: {ws / "dir"}
          dockerfile: {ws / "dir/Dockerfile_differ"}
        at "{ws / "action.yml"}", line 10, column 5 with params:
          context: {ws / "dir"}
          dockerfile: {ws / "dir/Dockerfile"}
    """.rstrip()
    )


async def test_check_image_refs_unique_same_defs(
    batch_cl_factory: BatchClFactory,
) -> None:
    async with batch_cl_factory("batch_images") as cl:
        flow = await RunningBatchFlow.create(cl, "duplicate_ref_same", "bake-id")
        # Should not raise an exception
        await check_image_refs_unique(flow)


async def test_upload_image_data(
    batch_cl_factory: BatchClFactory,
    batch_storage: Storage,
    assets: pathlib.Path,
) -> None:
    async with batch_cl_factory("batch_images") as cl:
        flow = await RunningBatchFlow.create(cl, "batch", "bake-id")
        project = await batch_storage.get_or_create_project("test")
        project_storage = batch_storage.project(id=project.id)
        bake = await project_storage.create_bake(
            batch="batch",
            params=None,
            name=None,
            graphs={},
            tags=[],
        )
        bake_storage = project_storage.bake(id=bake.id)

        runs = []

        async def _fake_run_cli(*args: str) -> None:
            runs.append(args)

        await upload_image_data(flow, _fake_run_cli, bake_storage)

        ref2img: Mapping[str, BakeImage] = {
            image.ref: image async for image in bake_storage.list_bake_images()
        }
        assert ref2img.keys() == {"image:main", "image:banana1", "image:banana2"}
        for ref in {"image:main", "image:banana1"}:
            img = ref2img[ref]
            assert img.context_on_storage
            assert img.dockerfile_rel == "Dockerfile"
            assert any(
                run == ("mkdir", "--parents", str(img.context_on_storage))
                for run in runs
            )
            assert any(
                run
                == (
                    "cp",
                    "--recursive",
                    "--update",
                    "--no-target-directory",
                    str(assets / "batch_images/dir"),
                    str(img.context_on_storage),
                )
                for run in runs
            )



# test functions available in expressions
import pathlib
import pytest
import sys
from neuro_sdk import Client
from re_assert import Matches
from typing import AbstractSet, Any, AsyncIterator, Callable, Mapping

from neuro_flow.context import FlowCtx, LiveContext, ProjectCtx, TagsCtx
from neuro_flow.expr import EvalError, RootABC, StrExpr, TypeT
from neuro_flow.tokenizer import Pos
from neuro_flow.types import LocalPath


if sys.version_info >= (3, 7):  # pragma: no cover
    from contextlib import asynccontextmanager
else:
    from async_generator import asynccontextmanager


POS = Pos(0, 0, LocalPath(__file__))


class Root(RootABC):
    def __init__(self, mapping: Mapping[str, TypeT], client: Client) -> None:
        self._mapping = mapping
        self._client = client

    def lookup(self, name: str) -> TypeT:
        return self._mapping[name]

    @asynccontextmanager
    async def client(self) -> AsyncIterator[Client]:
        yield self._client


async def test_len(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ len('abc') }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "3"


@pytest.mark.parametrize(
    "pattern,result",
    [
        ("'just string'", "just string"),
        ("True", "True"),
        ("222", "222"),
        ("22.22", "22.22"),
        ("[1, 2, 3]", "[1, 2, 3]"),
    ],
)
async def test_str(client: Client, pattern: str, result: str) -> None:
    expr = StrExpr(POS, POS, "${{ str(" + pattern + ") }}")
    ret = await expr.eval(Root({}, client))
    assert ret == result


async def test_replace(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ replace('22.22', '.', '_') }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "22_22"


async def test_join(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ join('_', ['x', 'y', 'z']) }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "x_y_z"


async def test_keys(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ keys(dct) }}")
    ret = await expr.eval(Root({"dct": {"a": 1, "b": 2}}, client))
    assert ret == "['a', 'b']"


async def test_values(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ values(dct) }}")
    ret = await expr.eval(Root({"dct": {"a": 1, "b": 2}}, client))
    assert ret == "[1, 2]"


async def test_fmt(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ fmt('{} {}', 1, 'a') }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "1 a"


async def test_hash_files(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ hash_files('Dockerfile', 'requirements/*.txt') }}")
    folder = LocalPath(__file__).parent / "hash_files"
    ret = await expr.eval(Root({"flow": {"workspace": folder}}, client))
    assert ret == "081fde04651e1184890a0470501bff3db8e0014260224e07acf5688e70e0edbe"


async def test_lower(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ lower('aBcDeF') }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "abcdef"


async def test_upper(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ upper('aBcDeF') }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "ABCDEF"


async def test_parse_volume_id(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ parse_volume('storage:path/to:/mnt/path:rw').id }}")
    ret = await expr.eval(Root({}, client))
    assert ret == "<volume>"


async def test_parse_volume_remote(client: Client) -> None:
    expr = StrExpr(
        POS, POS, "${{ parse_volume('storage:path/to:/mnt/path:rw').remote }}"
    )
    ret = await expr.eval(Root({}, client))
    assert Matches("storage:[^:]+/path/to") == ret


async def test_parse_volume_mount(client: Client) -> None:
    expr = StrExpr(
        POS, POS, "${{ parse_volume('storage:path/to:/mnt/path:rw').mount }}"
    )
    ret = await expr.eval(Root({}, client))
    assert ret == "/mnt/path"


async def test_parse_volume_read_only(client: Client) -> None:
    expr = StrExpr(
        POS, POS, "${{ parse_volume('storage:path/to:/mnt/path:rw').read_only }}"
    )
    ret = await expr.eval(Root({}, client))
    assert ret == "False"


async def test_parse_volume_local(client: Client) -> None:
    expr = StrExpr(
        POS, POS, "${{ parse_volume('storage:path/to:/mnt/path:rw').local }}"
    )
    ret = await expr.eval(Root({}, client))
    assert ret == "None"


async def test_parse_volume_local_full(client: Client) -> None:
    expr = StrExpr(
        POS, POS, "${{ parse_volume('storage:path/to:/mnt/path:rw').full_local_path }}"
    )
    ret = await expr.eval(Root({}, client))
    assert ret == "None"


LiveContextFactory = Callable[[Client, TagsCtx], LiveContext]


@pytest.fixture
async def live_context_factory(assets: pathlib.Path) -> LiveContextFactory:
    def _factory(client: Client, tags: TagsCtx = frozenset()) -> LiveContext:
        return LiveContext(
            _client=client,
            project=ProjectCtx(
                id="test",
                owner=None,
                role=None,
            ),
            flow=FlowCtx(
                flow_id="live",
                project_id="test",
                workspace=assets,
                title="unit test flow",
            ),
            env={},
            images={},
            volumes={},
            tags=tags,
        )

    return _factory


async def test_inspect_job_bad_context(client: Client) -> None:
    expr = StrExpr(POS, POS, "${{ inspect_job(22) }}")
    with pytest.raises(
        EvalError, match=r"inspect_job\(\) is only available inside a job definition"
    ):
        await expr.eval(Root({}, client))


async def test_inspect_job_bad_first_arg(
    client: Client, live_context_factory: LiveContextFactory
) -> None:
    expr = StrExpr(POS, POS, "${{ inspect_job(22) }}")
    ctx = live_context_factory(client, set())
    with pytest.raises(
        EvalError, match=r"inspect_job\(\) job_name argument should be a str, got 22"
    ):
        await expr.eval(ctx)


async def test_inspect_job_bad_second_arg(
    client: Client, live_context_factory: LiveContextFactory
) -> None:
    expr = StrExpr(POS, POS, "${{ inspect_job('foo', 22) }}")
    ctx = live_context_factory(client, set())
    with pytest.raises(
        EvalError, match=r"inspect_job\(\) suffix argument should be a str, got 22"
    ):
        await expr.eval(ctx)


async def test_inspect_job_correct_set_of_tags(
    client: Client, live_context_factory: LiveContextFactory
) -> None:
    called = False

    async def fake_list(
        *args: Any, tags: AbstractSet[str], **kwargs: Any
    ) -> AsyncIterator[Any]:
        nonlocal called
        called = True
        assert tags == {"test_tag", "job:foo"}
        yield {"id": "test_id"}

    client.jobs.list = fake_list  # type: ignore

    expr = StrExpr(POS, POS, "${{ inspect_job('foo').id }}")

    ctx = live_context_factory(client, {"test_tag"})
    result = await expr.eval(ctx)
    assert result == "test_id"
    assert called


async def test_inspect_job_correct_set_of_tags_multi(
    client: Client, live_context_factory: LiveContextFactory
) -> None:
    called = False

    async def fake_list(
        *args: Any, tags: AbstractSet[str], **kwargs: Any
    ) -> AsyncIterator[Any]:
        nonlocal called
        called = True
        assert tags == {"test_tag", "job:foo", "multi:bar"}
        yield {"id": "test_id"}

    client.jobs.list = fake_list  # type: ignore

    expr = StrExpr(POS, POS, "${{ inspect_job('foo', 'bar').id }}")

    ctx = live_context_factory(client, {"test_tag"})
    result = await expr.eval(ctx)
    assert result == "test_id"
    assert called


async def test_inspect_job_no_jobs_errors(
    client: Client, live_context_factory: LiveContextFactory
) -> None:
    async def fake_list(*args: Any, **kwargs: Any) -> AsyncIterator[Any]:
        # make it async iterator:
        for _ in range(0):
            yield None

    client.jobs.list = fake_list  # type: ignore

    expr = StrExpr(POS, POS, "${{ inspect_job('foo') }}")
    expr_with_suffix = StrExpr(POS, POS, "${{ inspect_job('foo', 'bar') }}")

    ctx = live_context_factory(client, set())
    with pytest.raises(
        EvalError, match=r"inspect_job\(\) did not found running job with name foo"
    ):
        await expr.eval(ctx)

    with pytest.raises(
        EvalError,
        match=r"inspect_job\(\) did not found running job with name foo"
        r" and suffix bar",
    ):
        await expr_with_suffix.eval(ctx)



import pathlib

from neuro_flow import ast
from neuro_flow.expr import (
    MappingItemsExpr,
    OptBoolExpr,
    OptIntExpr,
    OptLocalPathExpr,
    OptRemotePathExpr,
    OptStrExpr,
    OptTimeDeltaExpr,
    RemotePathExpr,
    SequenceItemsExpr,
    SimpleIdExpr,
    SimpleOptStrExpr,
    StrExpr,
    URIExpr,
)
from neuro_flow.parser import parse_project_stream
from neuro_flow.tokenizer import Pos


def test_parse_full(assets: pathlib.Path) -> None:
    config_file = assets / "with_project_yaml" / "project.yml"
    with config_file.open() as stream:
        project = parse_project_stream(stream)
    assert project == ast.Project(
        Pos(0, 0, config_file),
        Pos(46, 0, config_file),
        id=SimpleIdExpr(
            Pos(0, 0, config_file),
            Pos(0, 0, config_file),
            "test_project",
        ),
        owner=SimpleOptStrExpr(
            Pos(0, 0, config_file),
            Pos(0, 0, config_file),
            "test-owner",
        ),
        role=SimpleOptStrExpr(
            Pos(0, 0, config_file),
            Pos(0, 0, config_file),
            "test-owner/roles/test-role",
        ),
        images={
            "image_a": ast.Image(
                Pos(5, 4, config_file),
                Pos(17, 0, config_file),
                ref=StrExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "image:banana"
                ),
                context=OptStrExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "dir"
                ),
                dockerfile=OptStrExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "dir/Dockerfile"
                ),
                build_args=SequenceItemsExpr(
                    [
                        StrExpr(
                            Pos(0, 0, config_file), Pos(0, 0, config_file), "--arg1"
                        ),
                        StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "val1"),
                        StrExpr(
                            Pos(0, 0, config_file),
                            Pos(0, 0, config_file),
                            "--arg2=val2",
                        ),
                    ]
                ),
                env=MappingItemsExpr(
                    {
                        "SECRET_ENV": StrExpr(
                            Pos(0, 0, config_file), Pos(0, 0, config_file), "secret:key"
                        ),
                    }
                ),
                volumes=SequenceItemsExpr(
                    [
                        OptStrExpr(
                            Pos(0, 0, config_file),
                            Pos(0, 0, config_file),
                            "secret:key:/var/secret/key.txt",
                        ),
                    ]
                ),
                build_preset=OptStrExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "gpu-small"
                ),
                force_rebuild=OptBoolExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
            )
        },
        volumes={
            "volume_a": ast.Volume(
                Pos(19, 4, config_file),
                Pos(23, 2, config_file),
                remote=URIExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "storage:dir"
                ),
                mount=RemotePathExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "/var/dir"
                ),
                read_only=OptBoolExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), True
                ),
                local=OptLocalPathExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "dir"
                ),
            ),
            "volume_b": ast.Volume(
                Pos(24, 4, config_file),
                Pos(26, 0, config_file),
                remote=URIExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "storage:other"
                ),
                mount=RemotePathExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "/var/other"
                ),
                read_only=OptBoolExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                local=OptLocalPathExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
            ),
        },
        defaults=ast.BatchFlowDefaults(
            Pos(27, 2, config_file),
            Pos(42, 0, config_file),
            _specified_fields={
                "fail_fast",
                "tags",
                "cache",
                "env",
                "volumes",
                "life_span",
                "schedule_timeout",
                "workdir",
                "max_parallel",
                "preset",
            },
            tags=SequenceItemsExpr(
                [
                    StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "tag-a"),
                    StrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), "tag-b"),
                ]
            ),
            env=MappingItemsExpr(
                {
                    "global_a": StrExpr(
                        Pos(0, 0, config_file), Pos(0, 0, config_file), "val-a"
                    ),
                    "global_b": StrExpr(
                        Pos(0, 0, config_file), Pos(0, 0, config_file), "val-b"
                    ),
                }
            ),
            volumes=SequenceItemsExpr(
                [
                    OptStrExpr(
                        Pos(0, 0, config_file),
                        Pos(0, 0, config_file),
                        "storage:common:/mnt/common:rw",
                    ),
                ]
            ),
            workdir=OptRemotePathExpr(
                Pos(0, 0, config_file), Pos(0, 0, config_file), "/global/dir"
            ),
            life_span=OptTimeDeltaExpr(
                Pos(0, 0, config_file), Pos(0, 0, config_file), "1d4h"
            ),
            preset=OptStrExpr(
                Pos(0, 0, config_file), Pos(0, 0, config_file), "cpu-large"
            ),
            schedule_timeout=OptTimeDeltaExpr(
                Pos(0, 0, config_file), Pos(0, 0, config_file), "24d23h22m21s"
            ),
            cache=ast.Cache(
                Pos(40, 4, config_file),
                Pos(42, 0, config_file),
                strategy=ast.CacheStrategy.NONE,
                life_span=OptTimeDeltaExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "2h30m"
                ),
            ),
            fail_fast=OptBoolExpr(
                Pos(0, 0, config_file), Pos(0, 0, config_file), False
            ),
            max_parallel=OptIntExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), 20),
        ),
        mixins={
            "basic": ast.ExecUnitMixin(
                Pos(44, 4, config_file),
                Pos(46, 0, config_file),
                _specified_fields={"image", "preset"},
                mixins=None,
                name=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                image=OptStrExpr(
                    Pos(0, 0, config_file),
                    Pos(0, 0, config_file),
                    "mixin-image",
                ),
                preset=OptStrExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), "mixin-preset"
                ),
                schedule_timeout=OptTimeDeltaExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                entrypoint=OptStrExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                cmd=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                workdir=OptRemotePathExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                env=None,
                volumes=None,
                tags=None,
                life_span=OptTimeDeltaExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                title=OptStrExpr(Pos(0, 0, config_file), Pos(0, 0, config_file), None),
                http_port=OptIntExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                http_auth=OptBoolExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
                pass_config=OptBoolExpr(
                    Pos(0, 0, config_file), Pos(0, 0, config_file), None
                ),
            ),
        },
    )



import json
import pathlib
import secrets
from yarl import URL

from tests.e2e.conftest import RunCLI


async def test_live_context(
    ws: pathlib.Path,
    run_cli: RunCLI,
    run_neuro_cli: RunCLI,
    project_id: str,
    username: str,
    project_role: str,
    cluster_name: str,
) -> None:
    captured = await run_cli(["run", "job_ctx_full", "--param", "arg1", "cli-value"])

    DUMP_START_MARK = "DUMP_START"
    DUMP_START_END = "DUMP_END"
    mark_found = False
    json_str = ""

    for line in captured.out.splitlines():
        if line == DUMP_START_MARK:
            mark_found = True
        elif line == DUMP_START_END:
            break
        elif mark_found:
            json_str += line
    result = json.loads(json_str)
    job_id = result.pop("job-id")
    assert result == {
        "project": {
            "id": project_id,
            "owner": username,
            "role": project_role,
        },
        "flow": {
            "flow_id": "live",
            "project_id": project_id,
            "workspace": str(ws),
            "title": "Test live config",
        },
        "env": {"FOO": "FOO_ENV"},
        "tags": ["flow:live", f"project:{project_id.replace('_', '-')}"],
        "volumes": {
            "ro_dir": {
                "id": "ro_dir",
                "remote": f"storage:neuro-flow-e2e/{project_id}/ro_dir",
                "mount": "/project/ro_dir",
                "read_only": True,
                "local": "ro_dir",
                "full_local_path": str(ws / "ro_dir"),
            },
            "rw_dir": {
                "id": "rw_dir",
                "remote": f"storage:neuro-flow-e2e/{project_id}/rw_dir",
                "mount": "/project/rw_dir",
                "read_only": False,
                "local": "rw_dir",
                "full_local_path": str(ws / "rw_dir"),
            },
        },
        "images": {
            "img": {
                "id": "img",
                "ref": f"image:neuro-flow-e2e/{project_id}:v1.0",
                "context": str(ws),
                "dockerfile": str(ws / "Dockerfile"),
                "dockerfile_rel": "Dockerfile",
                "build_preset": None,
                "build_args": [],
                "env": {},
                "volumes": [],
                "force_rebuild": False,
            }
        },
        "params": {"arg1": "cli-value", "arg2": "val2"},
    }

    job_uri = URL.build(scheme="job", host=cluster_name) / username / job_id
    captured = await run_neuro_cli(["acl", "list", "--shared", str(job_uri)])
    assert sorted(line.split() for line in captured.out.splitlines()) == [
        [f"job:{job_id}", "write", project_role],
    ]


async def test_volumes(
    ws: pathlib.Path,
    run_cli: RunCLI,
    run_neuro_cli: RunCLI,
    project_id: str,
    username: str,
    project_role: str,
    cluster_name: str,
) -> None:
    await run_cli(["mkvolumes"])

    storage_uri = URL.build(scheme="storage", host=cluster_name)
    storage_uri = storage_uri / username / "neuro-flow-e2e" / project_id
    captured = await run_neuro_cli(["acl", "list", "--shared", str(storage_uri)])
    assert sorted(line.split() for line in captured.out.splitlines()) == [
        [f"storage:neuro-flow-e2e/{project_id}/ro_dir", "write", project_role],
        [f"storage:neuro-flow-e2e/{project_id}/rw_dir", "write", project_role],
    ]

    await run_cli(["upload", "ALL"])
    random_text = secrets.token_hex(20)
    (ws / "ro_dir/updated_file").write_text(random_text)
    captured = await run_cli(["run", "volumes_test"])
    assert "initial_file_content" in captured.out
    assert random_text in captured.out


async def test_image_build(
    run_cli: RunCLI,
    run_neuro_cli: RunCLI,
    project_id: str,
    username: str,
    project_role: str,
    cluster_name: str,
) -> None:
    image_uri = URL.build(scheme="image", host=cluster_name)
    image_uri = image_uri / username / "neuro-flow-e2e" / project_id
    try:
        await run_cli(["build", "img"])

        captured = await run_neuro_cli(["acl", "list", "--shared", str(image_uri)])
        assert sorted(line.split() for line in captured.out.splitlines()) == [
            [f"image:neuro-flow-e2e/{project_id}", "write", project_role],
        ]

        random_text = secrets.token_hex(20)
        captured = await run_cli(
            [
                "run",
                "image_alpine",
                "--param",
                "cmd",
                f"echo '{random_text}' && sleep 5",
            ]
        )
        assert random_text in captured.out
    finally:
        await run_neuro_cli(["image", "rm", str(image_uri)])



import operator
import pytest
from funcparserlib.parser import NoParseError
from textwrap import dedent
from typing import Any
from typing_extensions import Final

from neuro_flow.expr import (
    FUNCTIONS,
    PARSER,
    AttrGetter,
    BinOp,
    Call,
    DictMaker,
    ItemGetter,
    ListMaker,
    Literal,
    Lookup,
    Text,
    UnaryOp,
    logical_and,
    logical_or,
)
from neuro_flow.tokenizer import LexerError, Pos, tokenize
from neuro_flow.types import LocalPath


FNAME = LocalPath("<test>")
START: Final = Pos(0, 0, FNAME)


def test_tmpl_ok1() -> None:
    assert [Lookup(Pos(0, 4, FNAME), Pos(0, 8, FNAME), "name", [])] == PARSER.parse(
        list(tokenize("${{ name }}", START))
    )


def test_tmpl_ok2() -> None:
    assert [
        Lookup(
            Pos(0, 4, FNAME),
            Pos(0, 18, FNAME),
            "name",
            [
                AttrGetter(Pos(0, 9, FNAME), Pos(0, 12, FNAME), "sub"),
                AttrGetter(Pos(0, 13, FNAME), Pos(0, 18, FNAME), "param"),
            ],
        )
    ] == PARSER.parse(list(tokenize("${{ name.sub.param }}", START)))


def test_tmpl_ok3() -> None:
    assert [Lookup(Pos(0, 3, FNAME), Pos(0, 7, FNAME), "name", [])] == PARSER.parse(
        list(tokenize("${{name}}", START))
    )


def test_tmpl_ok4() -> None:
    assert [Lookup(Pos(0, 4, FNAME), Pos(0, 8, FNAME), "name", [])] == PARSER.parse(
        list(tokenize("$[[ name ]]", START))
    )


def test_tmpl_ok5() -> None:
    assert [
        Lookup(
            Pos(0, 4, FNAME),
            Pos(0, 18, FNAME),
            "name",
            [
                AttrGetter(Pos(0, 9, FNAME), Pos(0, 12, FNAME), "sub"),
                AttrGetter(Pos(0, 13, FNAME), Pos(0, 18, FNAME), "param"),
            ],
        )
    ] == PARSER.parse(list(tokenize("$[[ name.sub.param ]]", START)))


def test_tmpl_ok6() -> None:
    assert [Lookup(Pos(0, 3, FNAME), Pos(0, 7, FNAME), "name", [])] == PARSER.parse(
        list(tokenize("$[[name]]", START))
    )


def test_tmpl_false1() -> None:
    with pytest.raises(LexerError):
        PARSER.parse(list(tokenize("}}", START)))


def test_tmpl_false2() -> None:
    with pytest.raises(NoParseError):
        PARSER.parse(list(tokenize("${{", START)))


def test_tmpl_false3() -> None:
    with pytest.raises(NoParseError):
        PARSER.parse(list(tokenize("${{ name sub  param", START)))


def test_tmpl_false4() -> None:
    with pytest.raises(NoParseError):
        PARSER.parse(list(tokenize("${{ name", START)))


def test_tmpl_false5() -> None:
    with pytest.raises(LexerError):
        PARSER.parse(list(tokenize("name }}", START)))


def test_tmpl_false6() -> None:
    with pytest.raises(LexerError):
        PARSER.parse(list(tokenize("]]", START)))


def test_tmpl_false7() -> None:
    with pytest.raises(NoParseError):
        PARSER.parse(list(tokenize("$[[", START)))


def test_tmpl_false8() -> None:
    with pytest.raises(NoParseError):
        PARSER.parse(list(tokenize("$[[ name sub  param", START)))


def test_tmpl_false9() -> None:
    with pytest.raises(NoParseError):
        PARSER.parse(list(tokenize("$[[ name", START)))


def test_tmpl_false10() -> None:
    with pytest.raises(LexerError):
        PARSER.parse(list(tokenize("name ]]", START)))


def test_tmpl_literal_none() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 8, FNAME), None)] == PARSER.parse(
        list(tokenize("${{ None }}", START))
    )


def test_tmpl_literal_real() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 9, FNAME), 12.34)] == PARSER.parse(
        list(tokenize("${{ 12.34 }}", START))
    )


def test_tmpl_literal_exp() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 14, FNAME), -12.34e-21)] == PARSER.parse(
        list(tokenize("${{ -12.34e-21 }}", START))
    )


def test_tmpl_literal_int1() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 8, FNAME), 1234)] == PARSER.parse(
        list(tokenize("${{ 1234 }}", START))
    )


def test_tmpl_literal_int2() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 9, FNAME), 1234)] == PARSER.parse(
        list(tokenize("${{ 12_34 }}", START))
    )


def test_tmpl_literal_int3() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 9, FNAME), -1234)] == PARSER.parse(
        list(tokenize("${{ -1234 }}", START))
    )


def test_tmpl_literal_hex1() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 10, FNAME), 0x12AB)] == PARSER.parse(
        list(tokenize("${{ 0x12ab }}", START))
    )


def test_tmpl_literal_hex2() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 11, FNAME), 0x12AB)] == PARSER.parse(
        list(tokenize("${{ 0X12_ab }}", START))
    )


def test_tmpl_literal_oct1() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 10, FNAME), 0o1234)] == PARSER.parse(
        list(tokenize("${{ 0o1234 }}", START))
    )


def test_tmpl_literal_oct2() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 11, FNAME), 0o1234)] == PARSER.parse(
        list(tokenize("${{ 0O12_34 }}", START))
    )


def test_tmpl_literal_bin1() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 10, FNAME), 0b0110)] == PARSER.parse(
        list(tokenize("${{ 0b0110 }}", START))
    )


def test_tmpl_literal_bin2() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 11, FNAME), 0b0110)] == PARSER.parse(
        list(tokenize("${{ 0B01_10 }}", START))
    )


def test_tmpl_literal_bool1() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 8, FNAME), True)] == PARSER.parse(
        list(tokenize("${{ True }}", START))
    )


def test_tmpl_literal_bool2() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 9, FNAME), False)] == PARSER.parse(
        list(tokenize("${{ False }}", START))
    )


def test_tmpl_literal_str1() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 9, FNAME), "str")] == PARSER.parse(
        list(tokenize("${{ 'str' }}", START))
    )


def test_tmpl_literal_str2() -> None:
    assert [Literal(Pos(0, 4, FNAME), Pos(0, 13, FNAME), "abc\tdef")] == PARSER.parse(
        list(tokenize("${{ 'abc\tdef' }}", START))
    )


def test_text_ok() -> None:
    assert [Text(Pos(0, 0, FNAME), Pos(0, 9, FNAME), "some text")] == PARSER.parse(
        list(tokenize("some text", START))
    )


def test_text_with_dot() -> None:
    assert [Text(Pos(0, 0, FNAME), Pos(0, 11, FNAME), "some . text")] == PARSER.parse(
        list(tokenize("some . text", START))
    )


def test_parser1() -> None:
    assert [
        Text(Pos(0, 0, FNAME), Pos(0, 5, FNAME), "some "),
        Lookup(
            Pos(0, 9, FNAME),
            Pos(0, 16, FNAME),
            "var",
            [AttrGetter(Pos(0, 13, FNAME), Pos(0, 16, FNAME), "arg")],
        ),
        Text(Pos(0, 19, FNAME), Pos(0, 24, FNAME), " text"),
    ] == PARSER.parse(list(tokenize("some ${{ var.arg }} text", START)))


def test_func_call_empty() -> None:
    assert [
        Call(Pos(0, 4, FNAME), Pos(0, 11, FNAME), FUNCTIONS["nothing"], [], [])
    ] == PARSER.parse(list(tokenize("${{ nothing() }}", START)))


def test_func_call_single_arg() -> None:
    assert [
        Call(
            Pos(0, 4, FNAME),
            Pos(0, 13, FNAME),
            FUNCTIONS["len"],
            [Literal(Pos(0, 8, FNAME), Pos(0, 13, FNAME), "abc")],
            [],
        )
    ] == PARSER.parse(list(tokenize("${{ len('abc') }}", START)))


def test_func_nested_calls() -> None:
    assert [
        Call(
            Pos(0, 4, FNAME),
            Pos(0, 16, FNAME),
            FUNCTIONS["len"],
            [
                Call(
                    Pos(0, 8, FNAME),
                    Pos(0, 16, FNAME),
                    FUNCTIONS["keys"],
                    [Lookup(Pos(0, 13, FNAME), Pos(0, 16, FNAME), "abc", [])],
                    [],
                )
            ],
            [],
        )
    ] == PARSER.parse(list(tokenize("${{ len(keys(abc)) }}", START)))


def test_func_call_multiple_args() -> None:
    assert [
        Call(
            Pos(0, 4, FNAME),
            Pos(0, 27, FNAME),
            FUNCTIONS["fmt"],
            [
                Literal(Pos(0, 8, FNAME), Pos(0, 15, FNAME), "{} {}"),
                Literal(Pos(0, 17, FNAME), Pos(0, 22, FNAME), "abc"),
                Literal(Pos(0, 24, FNAME), Pos(0, 27, FNAME), 123),
            ],
            [],
        )
    ] == PARSER.parse(list(tokenize('${{ fmt("{} {}", "abc", 123) }}', START)))


def test_func_call_arg_lookup() -> None:
    assert [
        Call(
            Pos(0, 4, FNAME),
            Pos(0, 30, FNAME),
            FUNCTIONS["len"],
            [
                Lookup(
                    Pos(0, 8, FNAME),
                    Pos(0, 30, FNAME),
                    "images",
                    [
                        AttrGetter(Pos(0, 15, FNAME), Pos(0, 19, FNAME), "name"),
                        AttrGetter(Pos(0, 20, FNAME), Pos(0, 30, FNAME), "build_args"),
                    ],
                )
            ],
            [],
        )
    ] == PARSER.parse(list(tokenize("${{ len(images.name.build_args) }}", START)))


def test_func_call_with_trailer_attr() -> None:
    assert [
        Call(
            Pos(0, 4, FNAME),
            Pos(0, 39, FNAME),
            FUNCTIONS["from_json"],
            [Literal(Pos(0, 14, FNAME), Pos(0, 36, FNAME), '{"a": 1, "b": "val"}')],
            [AttrGetter(Pos(0, 38, FNAME), Pos(0, 39, FNAME), "a")],
        )
    ] == PARSER.parse(
        list(tokenize("""${{ from_json('{"a": 1, "b": "val"}').a }}""", START))
    )


def test_func_call_with_trailer_item() -> None:
    assert [
        Call(
            Pos(0, 4, FNAME),
            Pos(0, 41, FNAME),
            FUNCTIONS["from_json"],
            [Literal(Pos(0, 14, FNAME), Pos(0, 36, FNAME), '{"a": 1, "b": "val"}')],
            [
                ItemGetter(
                    Pos(0, 38, FNAME),
                    Pos(0, 41, FNAME),
                    Literal(Pos(0, 38, FNAME), Pos(0, 41, FNAME), "a"),
                )
            ],
        )
    ] == PARSER.parse(
        list(tokenize("""${{ from_json('{"a": 1, "b": "val"}')['a'] }}""", START))
    )


@pytest.mark.parametrize(
    "op_str,op_func",
    [
        ("==", operator.eq),
        ("!=", operator.ne),
        ("or", logical_or),
        ("and", logical_and),
        ("<", operator.lt),
        ("<=", operator.le),
        (">", operator.gt),
        (">=", operator.ge),
    ],
)
def test_operator_parse(op_str: str, op_func: Any) -> None:
    assert [
        BinOp(
            Pos(0, 4, FNAME),
            Pos(0, 14 + len(op_str), FNAME),
            op_func,
            Lookup(Pos(0, 4, FNAME), Pos(0, 7, FNAME), "foo", []),
            Literal(
                Pos(0, 9 + len(op_str), FNAME),
                Pos(0, 14 + len(op_str), FNAME),
                "bar",
            ),
        )
    ] == PARSER.parse(list(tokenize(f"""${{{{ foo {op_str} "bar" }}}}""", START)))


@pytest.mark.parametrize(
    "op_str,op_func",
    [
        ("not", operator.not_),
    ],
)
def test_unary_operator_parse(op_str: str, op_func: Any) -> None:
    assert [
        UnaryOp(
            Pos(0, 4, FNAME),
            Pos(0, 10 + len(op_str), FNAME),
            op_func,
            Literal(
                Pos(0, 5 + len(op_str), FNAME),
                Pos(0, 10 + len(op_str), FNAME),
                "bar",
            ),
        )
    ] == PARSER.parse(list(tokenize(f"""${{{{ {op_str} "bar" }}}}""", START)))


def test_operator_parse_brackets() -> None:
    assert [
        BinOp(
            start=Pos(line=0, col=4, filename=FNAME),
            end=Pos(line=0, col=24, filename=FNAME),
            op=logical_or,
            left=Lookup(
                start=Pos(line=0, col=4, filename=FNAME),
                end=Pos(line=0, col=7, filename=FNAME),
                root="foo",
                trailer=[],
            ),
            right=BinOp(
                start=Pos(line=0, col=12, filename=FNAME),
                end=Pos(line=0, col=24, filename=FNAME),
                op=operator.eq,
                left=Lookup(
                    start=Pos(line=0, col=12, filename=FNAME),
                    end=Pos(line=0, col=15, filename=FNAME),
                    root="bar",
                    trailer=[],
                ),
                right=Literal(
                    start=Pos(line=0, col=19, filename=FNAME),
                    end=Pos(line=0, col=24, filename=FNAME),
                    val="baz",
                ),
            ),
        )
    ] == PARSER.parse(list(tokenize("""${{ foo or (bar == "baz") }}""", START)))


def test_corner_case1() -> None:
    s = dedent(
        """\
            jupyter notebook
              --no-browser
              --ip=0.0.0.0
              --allow-root
              --NotebookApp.token=
              --notebook-dir=${{ volumes.notebooks.mount }}
        """
    )
    assert (
        [
            Text(
                Pos(0, 0, FNAME),
                Pos(5, 17, FNAME),
                dedent(
                    """\
                            jupyter notebook
                              --no-browser
                              --ip=0.0.0.0
                              --allow-root
                              --NotebookApp.token=
                              --notebook-dir="""
                ),
            ),
            Lookup(
                Pos(5, 21, FNAME),
                Pos(5, 44, FNAME),
                "volumes",
                [
                    AttrGetter(Pos(5, 29, FNAME), Pos(5, 38, FNAME), "notebooks"),
                    AttrGetter(Pos(5, 39, FNAME), Pos(5, 44, FNAME), "mount"),
                ],
            ),
            Text(Pos(5, 47, FNAME), Pos(6, 0, FNAME), "\n"),
        ]
        == PARSER.parse(list(tokenize(s, START)))
    )


def test_corner_case2() -> None:
    s = dedent(
        """\
            bash -c 'cd ${{ volumes.project.mount }} &&
              python -u ${{ volumes.code.mount }}/train.py
                --data ${{ volumes.data.mount }}'
        """
    )
    assert [
        Text(Pos(0, 0, FNAME), Pos(0, 12, FNAME), "bash -c 'cd "),
        Lookup(
            Pos(0, 16, FNAME),
            Pos(0, 37, FNAME),
            "volumes",
            [
                AttrGetter(Pos(0, 24, FNAME), Pos(0, 31, FNAME), "project"),
                AttrGetter(Pos(0, 32, FNAME), Pos(0, 37, FNAME), "mount"),
            ],
        ),
        Text(Pos(0, 40, FNAME), Pos(1, 12, FNAME), " &&\n  python -u "),
        Lookup(
            Pos(1, 16, FNAME),
            Pos(1, 34, FNAME),
            "volumes",
            [
                AttrGetter(Pos(1, 24, FNAME), Pos(1, 28, FNAME), "code"),
                AttrGetter(Pos(1, 29, FNAME), Pos(1, 34, FNAME), "mount"),
            ],
        ),
        Text(Pos(1, 37, FNAME), Pos(2, 11, FNAME), "/train.py\n    --data "),
        Lookup(
            Pos(2, 15, FNAME),
            Pos(2, 33, FNAME),
            "volumes",
            [
                AttrGetter(Pos(2, 23, FNAME), Pos(2, 27, FNAME), "data"),
                AttrGetter(Pos(2, 28, FNAME), Pos(2, 33, FNAME), "mount"),
            ],
        ),
        Text(Pos(2, 36, FNAME), Pos(3, 0, FNAME), "'\n"),
    ] == PARSER.parse(list(tokenize(s, START)))


def test_list() -> None:
    assert [
        ListMaker(
            start=Pos(0, 4, FNAME),
            end=Pos(0, 25, FNAME),
            items=[
                Literal(
                    start=Pos(0, 4, FNAME),
                    end=Pos(0, 5, FNAME),
                    val=1,
                ),
                Literal(
                    start=Pos(0, 7, FNAME),
                    end=Pos(0, 10, FNAME),
                    val="2",
                ),
                Literal(
                    start=Pos(0, 12, FNAME),
                    end=Pos(0, 16, FNAME),
                    val=True,
                ),
                Call(
                    start=Pos(0, 18, FNAME),
                    end=Pos(0, 25, FNAME),
                    func=FUNCTIONS["len"],
                    args=[
                        Lookup(
                            start=Pos(0, 22, FNAME),
                            end=Pos(0, 25, FNAME),
                            root="ctx",
                            trailer=[],
                        )
                    ],
                    trailer=[],
                ),
            ],
        )
    ] == PARSER.parse(list(tokenize("${{[1, '2', True, len(ctx)]}}", START)))


def test_dict() -> None:
    assert [
        DictMaker(
            start=Pos(0, 5, FNAME),
            end=Pos(0, 26, FNAME),
            items=[
                (
                    Literal(
                        start=Pos(0, 5, FNAME),
                        end=Pos(0, 6, FNAME),
                        val=1,
                    ),
                    Literal(
                        start=Pos(0, 8, FNAME),
                        end=Pos(0, 11, FNAME),
                        val="2",
                    ),
                ),
                (
                    Literal(
                        start=Pos(0, 13, FNAME),
                        end=Pos(0, 17, FNAME),
                        val=True,
                    ),
                    Call(
                        start=Pos(0, 19, FNAME),
                        end=Pos(0, 26, FNAME),
                        func=FUNCTIONS["len"],
                        args=[
                            Lookup(
                                start=Pos(0, 23, FNAME),
                                end=Pos(0, 26, FNAME),
                                root="ctx",
                                trailer=[],
                            )
                        ],
                        trailer=[],
                    ),
                ),
            ],
        )
    ] == PARSER.parse(list(tokenize("${{ {1: '2', True: len(ctx)} }}", START)))


def test_dict_short() -> None:
    assert [
        DictMaker(
            start=Pos(0, 5, FNAME),
            end=Pos(0, 11, FNAME),
            items=[
                (
                    Literal(
                        start=Pos(0, 5, FNAME),
                        end=Pos(0, 6, FNAME),
                        val=1,
                    ),
                    Literal(
                        start=Pos(0, 8, FNAME),
                        end=Pos(0, 11, FNAME),
                        val="2",
                    ),
                ),
            ],
        )
    ] == PARSER.parse(list(tokenize("${{ {1: '2'} }}", START)))



from dataclasses import dataclass

import aiohttp
import asyncio
import logging
import os
import pathlib
import pytest
import secrets
import shutil
import subprocess
from datetime import datetime, timedelta
from neuro_sdk import Config, get as api_get, login_with_token
from pathlib import Path
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
)
from yarl import URL

from neuro_flow.context import sanitize_name


NETWORK_TIMEOUT = 3 * 60.0
CLIENT_TIMEOUT = aiohttp.ClientTimeout(None, None, NETWORK_TIMEOUT, NETWORK_TIMEOUT)

log = logging.getLogger(__name__)


@pytest.fixture
def loop() -> Iterator[asyncio.AbstractEventLoop]:
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    yield loop
    loop.run_until_complete(loop.shutdown_asyncgens())
    loop.close()


@pytest.fixture
def assets() -> pathlib.Path:
    return pathlib.Path(__file__).parent / "assets"


@pytest.fixture
def project_id() -> str:
    return f"e2e_proj_{secrets.token_hex(6)}_{make_date_flag()}"


async def get_config(nmrc_path: Optional[Path]) -> Config:
    __tracebackhide__ = True
    async with api_get(timeout=CLIENT_TIMEOUT, path=nmrc_path) as client:
        return client.config


@pytest.fixture
async def api_config_data(api_config: Optional[pathlib.Path]) -> Config:
    return await get_config(api_config)


@pytest.fixture
def username(api_config_data: Config) -> str:
    return api_config_data.username


@pytest.fixture
def cluster_name(api_config_data: Config) -> str:
    return api_config_data.cluster_name


@pytest.fixture
async def project_role(
    project_id: str, username: str, run_neuro_cli: "RunCLI"
) -> AsyncIterator[str]:
    project_role = f"{username}/projects/{sanitize_name(project_id)}"
    try:
        yield project_role
    finally:
        await run_neuro_cli(["acl", "remove-role", project_role])


@pytest.fixture
def ws(
    assets: pathlib.Path,
    tmp_path_factory: Any,
    project_id: str,
    username: str,
) -> pathlib.Path:
    tmp_dir: pathlib.Path = tmp_path_factory.mktemp("proj-dir-parent")
    ws_dir = tmp_dir / project_id
    shutil.copytree(assets / "ws", ws_dir)
    project_data = f"id: {project_id}"
    if username:
        project_data += f'\nowner: "{username}"'
    (ws_dir / ".neuro" / "project.yml").write_text(project_data)
    return ws_dir


@pytest.fixture
async def api_config(tmp_path_factory: Any) -> AsyncIterator[Optional[pathlib.Path]]:
    e2e_test_token = os.environ.get("E2E_USER_TOKEN")
    if e2e_test_token:
        tmp_path = tmp_path_factory.mktemp("config")
        config_path = tmp_path / "conftest"
        await login_with_token(
            e2e_test_token,
            url=URL("https://dev.neu.ro/api/v1"),
            path=config_path,
        )
    else:
        config_path = None
    yield config_path


@dataclass(frozen=True)
class SysCap:
    out: str
    err: str


RunCLI = Callable[[List[str]], Awaitable[SysCap]]


@pytest.fixture
def _run_cli(
    loop: None, ws: pathlib.Path, api_config: Optional[pathlib.Path]
) -> RunCLI:
    async def _run(
        arguments: List[str],
    ) -> SysCap:
        if api_config:
            os.environ["NEUROMATION_CONFIG"] = str(api_config)
        proc = await asyncio.create_subprocess_exec(
            *arguments,
            cwd=ws,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        assert proc.stdout
        assert proc.stderr
        ret_code = await asyncio.wait_for(proc.wait(), timeout=600)
        out = (await proc.stdout.read()).decode(
            encoding="utf8",
            errors="replace",
        )
        err = (await proc.stderr.read()).decode(
            encoding="utf8",
            errors="replace",
        )
        if ret_code:
            log.error(f"Last stdout: '{out}'")
            log.error(f"Last stderr: '{err}'")
            raise subprocess.CalledProcessError(ret_code, arguments[0], out, err)
        return SysCap(out=out.strip(), err=err.strip())

    return _run


@pytest.fixture
def run_cli(_run_cli: RunCLI) -> RunCLI:
    return lambda args: _run_cli(["neuro-flow", "--show-traceback"] + args)


@pytest.fixture
def run_neuro_cli(_run_cli: RunCLI) -> RunCLI:
    return lambda args: _run_cli(["neuro", "--show-traceback"] + args)


DATETIME_FORMAT = "%Y%m%d%H%M"
DATETIME_SEP = "date"


def make_date_flag() -> str:
    time_str = datetime.now().strftime(DATETIME_FORMAT)
    return f"{DATETIME_SEP}{time_str}{DATETIME_SEP}"


@pytest.fixture(scope="session")
def _drop_once_flag() -> Dict[str, bool]:
    return {}


@pytest.fixture(autouse=True)
async def drop_old_test_images(
    run_neuro_cli: RunCLI, _drop_once_flag: Dict[str, bool]
) -> None:
    if _drop_once_flag.get("cleaned_images"):
        return

    async def _drop_iamge(image_str: str) -> None:
        image_str = image_str.strip()
        image_url = URL(image_str)
        image_name = image_url.parts[-1]
        try:
            _, time_str, _ = image_name.split(DATETIME_SEP)
            image_time = datetime.strptime(time_str, DATETIME_FORMAT)
            if datetime.now() - image_time < timedelta(days=1):
                return
            await run_neuro_cli(["image", "rm", image_str])
        except Exception:
            pass

    res: SysCap = await run_neuro_cli(["-q", "image", "ls", "--full-uri"])

    tasks = []
    for image_str in res.out.splitlines():
        tasks.append(asyncio.ensure_future(_drop_iamge(image_str)))

    if tasks:
        await asyncio.wait(tasks)

    _drop_once_flag["cleaned_images"] = True


@pytest.fixture(autouse=True)
async def drop_old_roles(
    run_neuro_cli: RunCLI, _drop_once_flag: Dict[str, bool], username: str
) -> None:
    if _drop_once_flag.get("cleaned_roles"):
        return

    res: SysCap = await run_neuro_cli(
        ["acl", "ls", "--shared", f"role://{username}/projects"]
    )

    async def _drop_role(project_str: str) -> None:
        role_str, _ = project_str.split(" ", 1)
        role_uri = URL(role_str)
        role_name = role_uri.parts[-1]
        try:
            _, time_str, _ = role_name.split(DATETIME_SEP)
            proj_time = datetime.strptime(time_str, DATETIME_FORMAT)
            if datetime.now() - proj_time < timedelta(hours=4):
                return
            await run_neuro_cli(
                ["acl", "remove-role", f"{username}/projects/{role_name}"]
            )
        except Exception:
            pass

    tasks = []
    for project in res.out.splitlines():
        tasks.append(asyncio.ensure_future(_drop_role(project)))

    if tasks:
        await asyncio.wait(tasks)

    _drop_once_flag["cleaned_roles"] = True



import dataclasses

import click
import datetime
import neuro_extras
from collections import defaultdict
from graphviz import Digraph
from neuro_cli import __version__ as cli_version
from neuro_sdk import Client, ResourceNotFound, __version__ as sdk_version
from operator import attrgetter
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from types import TracebackType
from typing import (
    AbstractSet,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from typing_extensions import AsyncContextManager, AsyncIterator
from yarl import URL

import neuro_flow

from . import ast
from .batch_executor import BatchExecutor, LocalsBatchExecutor, get_running_flow
from .colored_topo_sorter import ColoredTopoSorter
from .commands import CmdProcessor
from .config_loader import BatchLocalCL
from .context import (
    EMPTY_ROOT,
    EarlyBatch,
    EarlyLocalCall,
    ProjectCtx,
    RunningBatchFlow,
    setup_project_ctx,
)
from .expr import EvalError, MultiError
from .parser import ConfigDir
from .storage.base import Attempt, Bake, BakeImage, BakeStorage, ProjectStorage, Storage
from .types import FullID, LocalPath, TaskStatus
from .utils import (
    CommandRunner,
    GlobalOptions,
    encode_global_options,
    fmt_datetime,
    fmt_timedelta,
    make_cmd_exec,
)


EXECUTOR_IMAGE = f"ghcr.io/neuro-inc/neuro-flow:{neuro_flow.__version__}"


GRAPH_COLORS = {
    TaskStatus.PENDING: "skyblue",
    TaskStatus.RUNNING: "steelblue",
    TaskStatus.SUCCEEDED: "limegreen",
    TaskStatus.CANCELLED: "orange",
    TaskStatus.SKIPPED: "magenta",
    TaskStatus.CACHED: "yellowgreen",
    TaskStatus.FAILED: "orangered",
    TaskStatus.UNKNOWN: "crimson",
}


class BakeFailedError(Exception):
    def __init__(self, status: TaskStatus):
        self.status = status


async def iter_flows(top_flow: EarlyBatch) -> AsyncIterator[Tuple[FullID, EarlyBatch]]:
    to_check: List[Tuple[FullID, EarlyBatch]] = [((), top_flow)]
    while to_check:
        prefix, flow = to_check.pop(0)
        yield prefix, flow
        for tid in flow.graph:
            if await flow.is_action(tid):
                sub_flow = await flow.get_action_early(tid)
                to_check.append((prefix + (tid,), sub_flow))


async def check_no_cycles(top_flow: EarlyBatch) -> None:
    async for _, flow in iter_flows(top_flow):
        ColoredTopoSorter(flow.graph)


async def check_local_deps(top_flow: EarlyBatch) -> None:
    # This methods works in O(kn^3), where:
    # - n is number of tasks in the flow
    # - k is maximal depths of actions
    # This complexity is because:
    # For each task (n) for task's each dependency (n) and for each remote task (n)
    # do prefix check (k). Note that task are ofter have a few dependencies,
    # so in real cases one of those n effectively const.
    #
    # If performance becomes a problem, it can be replaced
    # with Trie (prefix tree) to reduce time complexity to O(kn^2)
    # (for each task (n) for each task's dependency (n) do Trie check (k))

    runs_on_remote: Set[FullID] = set()

    async for prefix, flow in iter_flows(top_flow):
        runs_on_remote.update(
            {prefix + (tid,) for tid in flow.graph if await flow.is_task(tid)}
        )

    def _is_prefix(item: FullID, prefix: FullID) -> bool:
        if len(item) < len(prefix):
            return False
        return all(x == y for (x, y) in zip(item, prefix))

    def _remote_deps(prefix: FullID, deps: Iterable[str]) -> Iterable[FullID]:
        return (
            remote
            for dep in deps
            for remote in runs_on_remote
            if _is_prefix(remote, prefix + (dep,))
        )

    async for prefix, flow in iter_flows(top_flow):
        early_locals = cast(
            AsyncIterator[EarlyLocalCall],
            (
                await flow.get_local_early(tid)
                for tid in flow.graph
                if await flow.is_local(tid)
            ),
        )
        with_bad_deps = (
            (early_local, remote)
            async for early_local in early_locals
            for remote in _remote_deps(prefix, early_local.needs)
        )
        async for early_local, remote in with_bad_deps:
            early_local_str = ".".join(prefix + (early_local.real_id,))
            remote_str = ".".join(remote)
            raise Exception(
                f"Local action '{early_local_str}' depends on remote "
                f"task '{remote_str}'. This is not supported because "
                "all local action should succeed before "
                "remote executor starts."
            )


async def check_expressions(top_flow: RunningBatchFlow) -> None:
    errors: List[EvalError] = []
    async for _, flow in iter_flows(top_flow):
        errors += flow.validate_expressions()
    if errors:
        raise MultiError(errors)


class ImageRefNotUniqueError(Exception):
    @dataclasses.dataclass
    class ImageInfo:
        context: Optional[Union[URL, LocalPath]]
        dockerfile: Optional[Union[URL, LocalPath]]
        ast: ast.Image

    def __init__(self, ref: str, images: Sequence[ImageInfo]) -> None:
        self._ref = ref
        self._images = images

    def __str__(self) -> str:
        return (
            f"Image with ref '{self._ref}' defined multiple times "
            f"with different attributes:\n"
            + "\n".join(
                f"at {EvalError.format_pos(image.ast._start)} with params:\n"
                f"  context: {image.context or '<empty>'}\n"
                f"  dockerfile: {image.dockerfile or '<empty>'}"
                for image in self._images
            )
        )


async def check_image_refs_unique(top_flow: RunningBatchFlow) -> None:
    _tmp: Dict[str, List[ImageRefNotUniqueError.ImageInfo]] = defaultdict(list)

    async for _, flow in iter_flows(top_flow):
        for image in flow.early_images.values():
            if image.ref.startswith("image:"):
                _tmp[image.ref].append(
                    ImageRefNotUniqueError.ImageInfo(
                        context=image.context,
                        dockerfile=image.dockerfile,
                        ast=flow.get_image_ast(image.id),
                    )
                )

    errors = []
    for ref, images in _tmp.items():
        contexts_differ = len({it.context for it in images}) > 1
        dockerfiles_differ = len({it.dockerfile for it in images}) > 1
        if contexts_differ or dockerfiles_differ:
            errors.append(ImageRefNotUniqueError(ref, images))
    if errors:
        raise MultiError(errors)


async def build_graphs(
    top_flow: RunningBatchFlow,
) -> Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]]:
    graphs = {}

    async for prefix, flow in iter_flows(top_flow):
        graphs[prefix] = {
            prefix + (key,): {prefix + (node,) for node in nodes}
            for key, nodes in flow.graph.items()
        }

    return graphs


async def upload_image_data(
    top_flow: RunningBatchFlow,
    neuro_runner: CommandRunner,
    storage: BakeStorage,
) -> List[BakeImage]:
    @dataclasses.dataclass
    class _TmpData:
        context_on_storage: Optional[URL]
        dockerfile_rel: Optional[str]
        yaml_defs: List[FullID]

    _tmp: Dict[str, _TmpData] = {}

    async for prefix, flow in iter_flows(top_flow):
        for image in flow.early_images.values():
            if isinstance(image.context, LocalPath):
                # Reusing image ref between bakes introduces
                # race condition anyway, so we can safely use it
                # as remote context dir name
                storage_context_dir: Optional[URL] = URL(
                    f"storage:.flow/{top_flow.project_id}/{image.ref.replace(':', '/')}"
                )
            else:
                storage_context_dir = image.context

            dockerfile_rel = None
            if image.dockerfile_rel:
                dockerfile_rel = str(image.dockerfile_rel.as_posix())

            prev_entry = _tmp.get(image.ref)
            if prev_entry is not None:
                # Validation is done before
                prev_entry.yaml_defs.append(prefix + (image.id,))
            else:
                if isinstance(image.context, LocalPath):
                    await neuro_runner(
                        "mkdir",
                        "--parents",
                        str(storage_context_dir),
                    )
                    await neuro_runner(
                        "cp",
                        "--recursive",
                        "--update",
                        "--no-target-directory",
                        str(image.context),
                        str(storage_context_dir),
                    )
                _tmp[image.ref] = _TmpData(
                    yaml_defs=[prefix + (image.id,)],
                    context_on_storage=storage_context_dir,
                    dockerfile_rel=dockerfile_rel,
                )

    return [
        await storage.create_bake_image(
            ref=ref,
            yaml_defs=entry.yaml_defs,
            context_on_storage=entry.context_on_storage,
            dockerfile_rel=entry.dockerfile_rel,
        )
        for ref, entry in _tmp.items()
    ]


class BatchRunner(AsyncContextManager["BatchRunner"]):
    def __init__(
        self,
        config_dir: ConfigDir,
        console: Console,
        client: Client,
        storage: Storage,
        global_options: GlobalOptions,
        run_neuro_cli: Optional[CommandRunner] = None,
    ) -> None:
        self._config_dir = config_dir
        self._console = console
        self._client = client
        self._storage = storage
        self._project_storage: Optional[ProjectStorage] = None
        self._config_loader: Optional[BatchLocalCL] = None
        self._project: Optional[ProjectCtx] = None
        self._run_neuro_cli = run_neuro_cli or make_cmd_exec(
            "neuro", global_options=encode_global_options(global_options)
        )
        self._global_options = global_options

    @property
    def project_id(self) -> str:
        assert self._project is not None
        return self._project.id

    @property
    def project_role(self) -> Optional[str]:
        assert self._project is not None
        return self._project.role

    @property
    def config_loader(self) -> BatchLocalCL:
        assert self._config_loader is not None
        return self._config_loader

    @property
    def storage(self) -> ProjectStorage:
        assert self._project_storage is not None
        return self._project_storage

    async def close(self) -> None:
        if self._config_loader is not None:
            await self._config_loader.close()

    async def __aenter__(self) -> "BatchRunner":
        self._config_loader = BatchLocalCL(self._config_dir, self._client)
        self._project = await setup_project_ctx(EMPTY_ROOT, self._config_loader)
        project = await self._storage.get_or_create_project(self._project.id)
        self._project_storage = self._storage.project(id=project.id)
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    # Next function is also used in tests:
    async def _setup_bake(
        self,
        batch_name: str,
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> Tuple[Bake, RunningBatchFlow]:
        # batch_name is a name of yaml config inside self._workspace / .neuro
        # folder without the file extension
        self._console.log(f"[bright_black]neuro_sdk=={sdk_version}")
        self._console.log(f"[bright_black]neuro_cli=={cli_version}")
        self._console.log(f"[bright_black]neuro-extras=={neuro_extras.__version__}")
        self._console.log(f"[bright_black]neuro-flow=={neuro_flow.__version__}")
        self._console.log(f"Use config file {self.config_loader.flow_path(batch_name)}")

        # Check that the yaml is parseable
        flow = await RunningBatchFlow.create(
            self.config_loader, batch_name, "fake-bake-id", params
        )

        for volume in flow.volumes.values():
            if volume.local is not None:
                # TODO: sync volumes if needed
                raise NotImplementedError("Volumes sync is not supported")

        await check_no_cycles(flow)
        await check_local_deps(flow)
        await check_expressions(flow)
        await check_image_refs_unique(flow)
        graphs = await build_graphs(flow)

        self._console.log(
            "Check config... [green]ok[/green]",
        )

        self._console.log("Create bake...")

        bake = await self.storage.create_bake(
            batch=batch_name,
            graphs=graphs,
            params=flow.params,
            name=name,
            tags=tags,
        )
        bake_storage = self.storage.bake(id=bake.id)
        config_meta = await self.config_loader.collect_configs(batch_name, bake_storage)
        await bake_storage.create_attempt(number=1, configs_meta=config_meta)

        self._console.log(
            f"Bake [b]{bake.name or bake.id}[/b] of "
            f"project [b]{self.project_id}[/b] is created"
        )
        self._console.log("Uploading image contexts/dockerfiles...")
        await upload_image_data(flow, self._run_neuro_cli, bake_storage)

        return bake, flow

    # Next function is also used in tests:
    async def _run_locals(
        self,
        bake_id: str,
    ) -> TaskStatus:
        async with LocalsBatchExecutor.create(
            self._console,
            bake_id,
            self._client,
            self._storage,
            project_role=self.project_role,
        ) as executor:
            return await executor.run()

    async def bake(
        self,
        batch_name: str,
        local_executor: bool = False,
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> None:
        self._console.print(
            Panel(f"[bright_blue]Bake [b]{batch_name}[/b]", padding=1),
            justify="center",
        )
        bake, flow = await self._setup_bake(batch_name, params, name, tags)
        await self._run_bake(bake, flow, local_executor)

    async def _run_bake(
        self,
        bake: Bake,
        flow: RunningBatchFlow,
        local_executor: bool,
    ) -> None:
        self._console.rule("Run local actions")
        locals_result = await self._run_locals(bake.id)
        if locals_result != TaskStatus.SUCCEEDED:
            return
        self._console.rule("Run main actions")
        if local_executor:
            self._console.log(f"[bright_black]Using local executor")
            await self.process(bake.id)
        else:
            self._console.log(f"[bright_black]Starting remote executor")
            if flow.life_span:
                life_span = fmt_timedelta(flow.life_span)
            else:
                life_span = "7d"
            run_args = [
                "run",
                "--pass-config",
                f"--life-span={life_span}",
                f"--tag=project:{self.project_id}",
                f"--tag=flow:{bake.batch}",
                f"--tag=bake_id:{bake.id}",
                f"--tag=remote_executor",
            ]
            project_role = self.project_role
            if project_role is not None:
                run_args.append(f"--share={project_role}")
            run_args += [
                EXECUTOR_IMAGE,
                "--",
                "neuro-flow",
                *encode_global_options(self._global_options),
                "--fake-workspace",
                "execute",
                bake.id,
            ]
            await self._run_neuro_cli(*run_args)

    async def process(
        self,
        bake_id: str,
    ) -> None:
        async with BatchExecutor.create(
            self._console,
            bake_id,
            self._client,
            self._storage,
            project_role=self.project_role,
        ) as executor:
            status = await executor.run()
            if status != TaskStatus.SUCCEEDED:
                raise BakeFailedError(status)

    def get_bakes(self) -> AsyncIterator[Bake]:
        return self.storage.list_bakes()

    async def get_bake_attempt(self, bake_id: str, *, attempt_no: int = -1) -> Attempt:
        return await self._storage.bake(id=bake_id).attempt(number=attempt_no).get()

    async def list_bakes(
        self,
        tags: AbstractSet[str] = frozenset(),
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        recent_first: bool = False,
    ) -> None:
        def _setup_table() -> Table:
            table = Table(box=box.MINIMAL_HEAVY_HEAD)
            table.add_column(
                "ID",
                style="bold",
                min_width=len("2021-07-03T10:54:20+00:00_9e1f7a") + 10,
            )
            table.add_column("NAME", min_width=8)
            table.add_column(
                "EXECUTOR", width=len("job-f6bd815b-3a3b-4ea1-b5ec-e8ab13678e3e")
            )
            table.add_column("STATUS", width=9)
            table.add_column("WHEN", min_width=10)
            table.show_edge = False
            return table

        header_table = _setup_table()
        self._console.print(header_table)

        async for bake in self.storage.list_bakes(
            tags=tags,
            since=since,
            until=until,
            recent_first=recent_first,
        ):
            if bake.last_attempt is None:
                self._console.print(
                    f"[yellow]Bake [b]{bake.id}[/b] is malformed, skipping"
                )
            else:
                row_table = _setup_table()
                row_table.show_header = False
                row_table.add_row(
                    bake.id,
                    bake.name or "",
                    bake.last_attempt.executor_id or "",
                    bake.last_attempt.result,
                    fmt_datetime(bake.last_attempt.created_at),
                )
                self._console.print(row_table)

    async def inspect(
        self,
        bake_id: str,
        *,
        attempt_no: int = -1,
        output: Optional[LocalPath] = None,
        save_dot: bool = False,
        save_pdf: bool = False,
        view_pdf: bool = False,
    ) -> None:

        bake_storage = self.storage.bake(id=bake_id)
        try:
            bake = await bake_storage.get()
        except ResourceNotFound:
            self._console.print("[yellow]Bake not found")
            self._console.print(
                f"Please make sure that the bake [b]{bake_id}[/b] and "
                f"project [b]{self.project_id}[/b] are correct."
            )
            exit(1)
            assert False, "unreachable"

        attempt_storage = bake_storage.attempt(number=attempt_no)
        attempt = await attempt_storage.get()

        self._console.print(f"[b]Bake id: {bake_id}[/b]")
        self._console.print(f"[b]Attempt #{attempt.number}[/b]", attempt.result)
        if attempt.executor_id:
            info = await self._client.jobs.status(attempt.executor_id)
            self._console.print(
                f"[b]Executor {attempt.executor_id}[/b]", TaskStatus(info.status)
            )

        task_table = Table(box=box.MINIMAL_HEAVY_HEAD)
        task_table.add_column("ID", style="bold")
        task_table.add_column("STATUS")
        task_table.add_column("RAW ID", style="bright_black")
        task_table.add_column("STARTED")
        task_table.add_column("FINISHED")

        tasks = [task async for task in attempt_storage.list_tasks()]

        for task in sorted(tasks, key=attrgetter("created_at")):
            task_table.add_row(
                ".".join(task.yaml_id),
                task.status,
                task.raw_id,
                fmt_datetime(task.created_at),
                fmt_datetime(task.finished_at),
            )

        self._console.print("Tasks:")
        self._console.print(task_table)

        image_table = Table(box=box.MINIMAL_HEAVY_HEAD)
        image_table.add_column("REF", style="bold")
        image_table.add_column("STATUS")
        image_table.add_column("BUILDER ID", style="bright_black")
        async for image in bake_storage.list_bake_images():
            image_table.add_row(
                image.ref,
                image.status,
                image.builder_job_id or "",
            )
        if image_table.rows:
            self._console.print("Images:")
            self._console.print(image_table)

        if output is None:
            output = LocalPath(f"{bake.id}_{attempt.number}").with_suffix(".gv")

        graphs = bake.graphs
        dot = Digraph(bake.batch, filename=str(output), strict=True, engine="dot")
        dot.attr(compound="true")
        dot.node_attr = {"style": "filled"}

        await self._subgraph(
            dot, graphs, (), {}, {task.yaml_id: task.status for task in tasks}
        )

        if save_dot:
            self._console.print(f"Saving file {dot.filename}")
            dot.save()
        if save_pdf:
            self._console.print(f"Rendering {dot.filename}.pdf")
            dot.render(view=view_pdf)
        elif view_pdf:
            self._console.print(f"Opening {dot.filename}.pdf")
            dot.view()

    async def _subgraph(
        self,
        dot: Digraph,
        graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]],
        prefix: FullID,
        anchors: Dict[str, str],
        statuses: Dict[FullID, TaskStatus],
    ) -> None:
        lhead: Optional[str]
        ltail: Optional[str]
        color: Optional[str]
        first = True
        graph = graphs[prefix]
        for task_id, deps in graph.items():
            tgt = ".".join(task_id)
            name = task_id[-1]

            if first:
                anchors[".".join(prefix)] = tgt
                first = False

            if task_id in statuses:
                color = GRAPH_COLORS.get(statuses[task_id])
            else:
                color = None

            if task_id in graphs:
                lhead = "cluster_" + tgt
                with dot.subgraph(name=lhead) as subdot:
                    subdot.attr(label=f"{name}")
                    subdot.attr(compound="true")
                    subdot.attr(color=color)
                    await self._subgraph(
                        subdot,
                        graphs,
                        task_id,
                        anchors,
                        statuses,
                    )
                tgt = anchors[tgt]
            else:
                dot.node(tgt, name, color=color)
                lhead = None

            for dep in deps:
                src = ".".join(dep)
                if src in anchors:
                    # src is a subgraph
                    ltail = "cluster_" + src
                    src = anchors[src]
                else:
                    ltail = None
                dot.edge(src, tgt, ltail=ltail, lhead=lhead)

    async def logs(
        self, bake_id: str, task_id: str, *, attempt_no: int = -1, raw: bool = False
    ) -> None:
        attempt_storage = self.storage.bake(id=bake_id).attempt(number=attempt_no)
        attempt = await attempt_storage.get()
        full_id = tuple(task_id.split("."))
        try:
            task = await attempt_storage.task(yaml_id=full_id).get()
        except ResourceNotFound:
            raise click.BadArgumentUsage(f"Unknown task {task_id}")

        if not task.status.is_finished:
            raise click.BadArgumentUsage(f"Task {task_id} is not finished")

        self._console.print(f"[b]Attempt #{attempt.number}[/b]", attempt.result)
        self._console.print(f"Task [b]{task_id}[/b]", task.status)

        if not task.raw_id:
            return

        if raw:
            async for chunk in self._client.jobs.monitor(task.raw_id):
                self._console.print(chunk.decode("utf-8", "replace"), end="")
        else:
            async with CmdProcessor() as proc:
                async for chunk in self._client.jobs.monitor(task.raw_id):
                    async for line in proc.feed_chunk(chunk):
                        self._console.print(line.decode("utf-8", "replace"), end="")
                async for line in proc.feed_eof():
                    self._console.print(line.decode("utf-8", "replace"), end="")

    async def cancel(self, bake_id: str, *, attempt_no: int = -1) -> None:
        attempt_storage = self.storage.bake(id=bake_id).attempt(number=attempt_no)
        attempt = await attempt_storage.get()
        if attempt.result.is_finished:
            raise click.BadArgumentUsage(
                f"Attempt #{attempt.number} of {attempt.bake_id} is already stopped."
            )
        await attempt_storage.update(result=TaskStatus.CANCELLED)
        self._console.print(
            f"[b]Attempt #{attempt.number}[/b] of bake "
            f"[b]{attempt.bake_id}[/b] was cancelled."
        )

    async def clear_cache(
        self, batch: Optional[str] = None, task_id: Optional[str] = None
    ) -> None:
        full_id: Optional[FullID] = None
        if task_id:
            full_id = tuple(task_id.split("."))
        await self.storage.delete_cache_entries(batch, full_id)

    async def restart(
        self,
        bake_id: str,
        *,
        attempt_no: int = -1,
        from_failed: bool = True,
        local_executor: bool = False,
    ) -> None:
        bake, flow = await self._restart(
            bake_id, attempt_no=attempt_no, from_failed=from_failed
        )
        await self._run_bake(bake, flow, local_executor)

    async def _restart(
        self,
        bake_id: str,
        *,
        attempt_no: int = -1,
        from_failed: bool = True,
    ) -> Tuple[Bake, RunningBatchFlow]:
        bake_storage = self.storage.bake(id=bake_id)
        bake = await bake_storage.get()
        if bake.last_attempt and attempt_no == -1:
            last_attempt = attempt = bake.last_attempt
        else:
            attempt = await bake_storage.attempt(number=attempt_no).get()
            last_attempt = await bake_storage.last_attempt().get()
        if not attempt.result.is_finished:
            raise click.BadArgumentUsage(
                f"Cannot re-run still running attempt #{attempt.number} "
                f"of {bake.id}."
            )
        if not last_attempt.result.is_finished:
            raise click.BadArgumentUsage(
                f"Cannot re-run bake when last attempt #{last_attempt.number} "
                f"of {bake.id} is still running."
            )
        if attempt.result == TaskStatus.SUCCEEDED and from_failed:
            raise click.BadArgumentUsage(
                f"Cannot re-run successful attempt #{attempt.number} "
                f"of {bake.id} with `--from-failed` flag set.\n"
                "Hint: Try adding --no-from-failed to restart bake from the beginning."
            )
        if attempt.number >= 99:
            raise click.BadArgumentUsage(
                f"Cannot re-run {bake.id}, the number of attempts exceeded."
            )
        new_attempt = await bake_storage.create_attempt(
            number=last_attempt.number + 1,
            configs_meta=attempt.configs_meta,
        )
        if from_failed:
            new_attempt_storage = bake_storage.attempt(id=new_attempt.id)
            graphs = bake.graphs
            handled = set()  # a set of succesfully finished and not cached tasks
            tasks = {
                task.yaml_id: task
                async for task in bake_storage.attempt(id=attempt.id).list_tasks()
            }
            for task in sorted(tasks.values(), key=attrgetter("created_at")):
                if task.status == TaskStatus.SUCCEEDED:
                    # should check deps to don't process post-actions with
                    # always() precondition
                    prefix = task.yaml_id[:-1]
                    graph = graphs[prefix]
                    deps = graph[task.yaml_id]
                    if not deps or all(dep in handled for dep in deps):
                        if (
                            prefix in tasks
                            and tasks[prefix].status != TaskStatus.SUCCEEDED
                        ):
                            # If action didn't succeeded, we should create task manually
                            await new_attempt_storage.create_task(
                                yaml_id=prefix,
                                status=TaskStatus.PENDING,
                                raw_id=None,
                            )
                        # TODO allow to create task with multiple statuses
                        # and copy them from old task
                        await new_attempt_storage.create_task(
                            yaml_id=task.yaml_id,
                            status=TaskStatus.SUCCEEDED,
                            raw_id=task.raw_id,
                            outputs=task.outputs,
                            state=task.state,
                        )
                        handled.add(task.id)

        self._console.print(f"[b]Attempt #{new_attempt.number}[/b] is created")
        flow = await get_running_flow(
            bake, self._client, bake_storage, new_attempt.configs_meta
        )

        return bake, flow



import re
import secrets
from yarl import URL

from tests.e2e.conftest import RunCLI


#  TODO: remove --local-executor when projoect.yml parsing is fixed


async def test_seq_batch(run_cli: RunCLI) -> None:
    random_text = secrets.token_hex(20)
    captured = await run_cli(
        ["bake", "--local-executor", "seq", "--param", "token", random_text]
    )

    assert f"task_b_out: {random_text}" in captured.out

    # Now test the cache:

    captured = await run_cli(
        ["bake", "--local-executor", "seq", "--param", "token", random_text]
    )

    assert "cached" in captured.out


async def test_batch_with_local(
    run_cli: RunCLI,
    run_neuro_cli: RunCLI,
    project_id: str,
    username: str,
    project_role: str,
    cluster_name: str,
) -> None:
    captured = await run_cli(
        [
            "bake",
            "--local-executor",
            "print-local",
            "--param",
            "file_path",
            "rw_dir/initial_file",
        ]
    )

    assert f"file_content: initial_file_content" in captured.out

    m = re.search(r"Task print_readme (job-\S+) is", captured.out)
    assert m is not None
    job_id = m[1]
    job_uri = URL.build(scheme="job", host=cluster_name) / username / job_id
    captured = await run_neuro_cli(["acl", "list", "--shared", str(job_uri)])
    assert sorted(line.split() for line in captured.out.splitlines()) == [
        [f"job:{job_id}", "write", project_role],
    ]


async def test_batch_action(run_cli: RunCLI) -> None:
    captured = await run_cli(["bake", "prime-checks", "--local-executor"])

    assert f"5 is prime" in captured.out
    assert f"4 is not prime" in captured.out
    assert f"3 is prime" in captured.out



from neuro_flow.context import (
    BatchActionContext,
    BatchContext,
    BatchTaskContext,
    EmptyRoot,
)
from neuro_flow.expr import StrExpr
from neuro_flow.expr_validation import validate_expr
from neuro_flow.tokenizer import Pos
from neuro_flow.types import LocalPath


def test_expr_validation_ok() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ flow.flow_id }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors == []


def test_expr_validation_ok_for_property_access() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ volumes.volume.ref_rw }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors == []


def test_expr_validation_unknown_context() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ something_new }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors[0].args[0] == "Unknown context 'something_new'"


def test_expr_validation_not_context_field() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ flow.foo }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors[0].args[0] == "'BatchFlowCtx' has no attribute 'foo'"


def test_expr_validation_invalid_need() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ needs.fff }}",
    )
    errors = validate_expr(expr, BatchTaskContext, known_needs=set())
    assert errors[0].args[0] == "Task 'fff' is not available under needs context"


def test_expr_validation_invalid_input() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ inputs.fff }}",
    )
    errors = validate_expr(expr, BatchActionContext[EmptyRoot], known_inputs=set())
    assert errors[0].args[0] == "Input 'fff' is undefined"


def test_expr_validation_bad_subcontext_lookup() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ needs.task.foo }}",
    )
    errors = validate_expr(expr, BatchTaskContext, known_needs={"task"})
    assert errors[0].args[0] == "'DepCtx' has no attribute 'foo'"


def test_expr_validation_bad_set_attr() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ tags.anything }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors[0].args[0] == "'TagsCtx' has no attribute 'anything'"


def test_expr_validation_bad_indexing() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ flow['flow_id'] }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors[0].args[0] == "'BatchFlowCtx' is not subscriptable"


def test_expr_validation_set_indexing() -> None:
    expr = StrExpr(
        Pos(0, 0, LocalPath("<default>")),
        Pos(0, 0, LocalPath("<default>")),
        pattern="${{ tags['flow_id'] }}",
    )
    errors = validate_expr(expr, BatchContext)
    assert errors[0].args[0] == "'TagsCtx' is not subscriptable"



from dataclasses import replace

import asyncio
import pytest
import shutil
import sys
from datetime import datetime, timedelta, timezone
from neuro_sdk import (
    Client,
    Container,
    DiskVolume,
    HTTPPort,
    JobDescription,
    JobRestartPolicy,
    JobStatus,
    JobStatusHistory,
    RemoteImage,
    ResourceNotFound,
    Resources,
    SecretFile,
    Tag,
    Volume,
    get as api_get,
)
from pathlib import Path
from rich import get_console
from tempfile import TemporaryDirectory
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
)
from yarl import URL

from neuro_flow.batch_executor import BatchExecutor, LocalsBatchExecutor
from neuro_flow.batch_runner import BatchRunner
from neuro_flow.parser import ConfigDir
from neuro_flow.storage.base import Storage, Task
from neuro_flow.storage.in_memory import InMemoryStorage
from neuro_flow.types import ImageStatus, LocalPath, TaskStatus


MakeBatchRunner = Callable[[Path], Awaitable[BatchRunner]]


def make_descr(
    job_id: str,
    *,
    status: JobStatus = JobStatus.PENDING,
    tags: Iterable[str] = (),
    description: str = "",
    scheduler_enabled: bool = False,
    created_at: datetime = datetime.now(timezone.utc),
    started_at: Optional[datetime] = None,
    finished_at: Optional[datetime] = None,
    exit_code: Optional[int] = None,
    restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER,
    life_span: float = 3600,
    name: Optional[str] = None,
    container: Optional[Container] = None,
    pass_config: bool = False,
) -> JobDescription:
    if container is None:
        container = Container(RemoteImage("ubuntu"), Resources(100, 0.1))

    return JobDescription(
        id=job_id,
        owner="test-user",
        cluster_name="default",
        status=status,
        history=JobStatusHistory(
            status=status,
            reason="",
            description="",
            restarts=0,
            created_at=created_at,
            started_at=started_at,
            finished_at=finished_at,
            exit_code=exit_code,
        ),
        container=container,
        scheduler_enabled=scheduler_enabled,
        uri=URL(f"job://default/test-user/{job_id}"),
        name=name,
        tags=sorted(
            list(set(tags) | {"project:test", "flow:batch-seq", f"task:{job_id}"})
        ),
        description=description,
        restart_policy=restart_policy,
        life_span=life_span,
        pass_config=pass_config,
    )


class JobsMock:
    # Mock for client.jobs subsystem

    _data: Dict[str, JobDescription]
    _outputs: Dict[str, bytes]
    exception_on_status: Optional[BaseException]

    def __init__(self) -> None:
        self.id_counter = 0
        self._data = {}
        self._outputs = {}
        self.exception_on_status = None

    def _make_next_id(self) -> int:
        self.id_counter += 1
        return self.id_counter

    # Method for controlling in tests

    async def _get_task(self, job_id_or_task_name: str) -> JobDescription:
        job_id = job_id_or_task_name
        task_name = job_id_or_task_name.replace("_", "-")
        while True:
            if job_id in self._data:
                return self._data[job_id]
            for descr in self._data.values():
                if f"task:{task_name}" in descr.tags:
                    return descr
            await asyncio.sleep(0.01)

    async def get_task(
        self, job_id_or_task_name: str, timeout: float = 2
    ) -> JobDescription:
        try:
            return await asyncio.wait_for(
                self._get_task(job_id_or_task_name), timeout=timeout
            )
        except asyncio.TimeoutError:
            raise AssertionError(
                f'Task "{job_id_or_task_name}" did not appeared in timeout of {timeout}'
            )

    async def mark_started(self, job_id_or_task_name: str) -> None:
        descr = await self.get_task(job_id_or_task_name)
        if descr.status == JobStatus.PENDING:
            descr = replace(descr, status=JobStatus.RUNNING)
            new_history = replace(
                descr.history,
                status=JobStatus.RUNNING,
                started_at=datetime.now(timezone.utc),
            )
            descr = replace(descr, history=new_history)
            self._data[descr.id] = descr

    async def mark_done(self, job_id_or_task_name: str, output: bytes = b"") -> None:
        await self.mark_started(job_id_or_task_name)
        descr = await self.get_task(job_id_or_task_name)
        if descr.status == JobStatus.RUNNING:
            descr = replace(descr, status=JobStatus.SUCCEEDED)
            new_history = replace(
                descr.history,
                status=JobStatus.SUCCEEDED,
                finished_at=datetime.now(timezone.utc),
                exit_code=0,
            )
            descr = replace(descr, history=new_history)
            self._data[descr.id] = descr
        self._outputs[descr.id] = output

    async def mark_failed(self, job_id_or_task_name: str, output: bytes = b"") -> None:
        await self.mark_started(job_id_or_task_name)
        descr = await self.get_task(job_id_or_task_name)
        if descr.status == JobStatus.RUNNING:
            descr = replace(descr, status=JobStatus.FAILED)
            new_history = replace(
                descr.history,
                status=JobStatus.FAILED,
                finished_at=datetime.now(timezone.utc),
                exit_code=255,
            )
            descr = replace(descr, history=new_history)
            self._data[descr.id] = descr
        self._outputs[descr.id] = output

    def clear(self) -> None:
        self._data = {}
        self._outputs = {}

    # Fake Jobs methods

    async def start(
        self,
        *,
        image: RemoteImage,
        preset_name: str,
        entrypoint: Optional[str] = None,
        command: Optional[str] = None,
        working_dir: Optional[str] = None,
        http: Optional[HTTPPort] = None,
        env: Optional[Mapping[str, str]] = None,
        volumes: Sequence[Volume] = (),
        secret_env: Optional[Mapping[str, URL]] = None,
        secret_files: Sequence[SecretFile] = (),
        disk_volumes: Sequence[DiskVolume] = (),
        tty: bool = False,
        shm: bool = False,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
        description: Optional[str] = None,
        pass_config: bool = False,
        wait_for_jobs_quota: bool = False,
        schedule_timeout: Optional[float] = None,
        restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER,
        life_span: Optional[float] = None,
        privileged: bool = False,
    ) -> JobDescription:
        job_id = f"job-{self._make_next_id()}"

        resources = Resources(cpu=1, memory_mb=1, shm=shm)
        container = Container(
            image=image,
            entrypoint=entrypoint,
            command=command,
            http=http,
            resources=resources,
            env=env or {},
            secret_env=secret_env or {},
            volumes=volumes,
            working_dir=working_dir,
            secret_files=secret_files,
            disk_volumes=disk_volumes,
            tty=tty,
        )
        self._data[job_id] = JobDescription(
            id=job_id,
            owner="test-user",
            cluster_name="default",
            preset_name=preset_name,
            status=JobStatus.PENDING,
            history=JobStatusHistory(
                status=JobStatus.PENDING,
                reason="",
                description="",
                created_at=datetime.now(timezone.utc),
                restarts=0,
            ),
            container=container,
            scheduler_enabled=False,
            uri=URL(f"job://default/test-user/{job_id}"),
            name=name,
            tags=tags,
            description=description,
            restart_policy=restart_policy,
            life_span=life_span,
            pass_config=pass_config,
            privileged=privileged,
            schedule_timeout=schedule_timeout,
        )
        return self._data[job_id]

    async def kill(self, job_id: str) -> None:
        descr = self._data[job_id]
        descr = replace(descr, status=JobStatus.CANCELLED)
        new_history = replace(
            descr.history,
            status=JobStatus.CANCELLED,
            started_at=descr.history.started_at or datetime.now(timezone.utc),
            finished_at=datetime.now(timezone.utc),
            exit_code=0,
        )
        descr = replace(descr, history=new_history)
        self._data[descr.id] = descr

    async def status(self, job_id: str) -> JobDescription:
        if self.exception_on_status:
            raise self.exception_on_status
        try:
            return self._data[job_id]
        except KeyError:
            raise ResourceNotFound

    async def monitor(self, job_id: str) -> AsyncIterator[bytes]:
        yield self._outputs.get(job_id, b"")


class ImagesMock:
    def __init__(self) -> None:
        self.known_images: Dict[str, Tag] = {}

    async def tag_info(self, remote: RemoteImage) -> Tag:
        tag = self.known_images.get(str(remote))
        if tag:
            return tag
        raise ResourceNotFound

    async def _close(self) -> None:
        pass


@pytest.fixture()
def batch_storage(loop: None) -> Storage:
    return InMemoryStorage()


class FakeGlobalOptions:
    verbosity = 0
    show_traceback = False


class MockCliRunner:
    def __init__(self) -> None:
        self.runs: List[Tuple[str, ...]] = []

    async def run(self, *args: str) -> None:
        self.runs.append(args)


@pytest.fixture()
async def mock_neuro_cli_runner() -> MockCliRunner:
    return MockCliRunner()


class MockBuilder:
    def __init__(self, jobs_mock: JobsMock) -> None:
        self.runs: List[Tuple[str, ...]] = []
        self.jobs = jobs_mock
        self.ref2job: Dict[str, str] = {}

    async def run(self, *args: str) -> str:
        self.runs.append(args)
        job = await self.jobs.start(image=RemoteImage("test"), preset_name="fake")
        self.ref2job[args[-1]] = job.id
        return job.id


@pytest.fixture()
async def mock_builder(
    jobs_mock: JobsMock,
) -> MockBuilder:
    return MockBuilder(jobs_mock)


@pytest.fixture()
async def make_batch_runner(
    batch_storage: Storage,
    client: Client,
    mock_neuro_cli_runner: MockCliRunner,
) -> AsyncIterator[MakeBatchRunner]:

    runner: Optional[BatchRunner] = None

    async def create(path: Path) -> BatchRunner:
        config_dir = ConfigDir(
            workspace=path,
            config_dir=path,
        )
        nonlocal runner
        # BatchRunner should not use client in this case

        runner = BatchRunner(
            config_dir,
            get_console(),
            client,
            batch_storage,
            FakeGlobalOptions(),
            run_neuro_cli=mock_neuro_cli_runner.run,
        )
        await runner.__aenter__()
        return runner

    yield create

    if runner is not None:
        await runner.__aexit__(None, None, None)


@pytest.fixture()
async def batch_runner(
    make_batch_runner: MakeBatchRunner,
    assets: Path,
) -> BatchRunner:
    return await make_batch_runner(assets)


@pytest.fixture()
def jobs_mock() -> JobsMock:
    return JobsMock()


@pytest.fixture()
def images_mock() -> ImagesMock:
    return ImagesMock()


@pytest.fixture()
async def patched_client(
    api_config: Path, jobs_mock: JobsMock, images_mock: ImagesMock
) -> AsyncIterator[Client]:
    async with api_get(path=api_config) as client:
        client._jobs = jobs_mock
        client._images = images_mock  # type: ignore
        yield client


@pytest.fixture()
def start_locals_executor(
    batch_storage: Storage, patched_client: Client
) -> Callable[[str], Awaitable[TaskStatus]]:
    async def start(bake_id: str) -> TaskStatus:
        async with LocalsBatchExecutor.create(
            get_console(),
            bake_id,
            patched_client,
            batch_storage,
        ) as executor:
            return await executor.run()

    return start


@pytest.fixture()
def start_executor(
    batch_storage: Storage,
    patched_client: Client,
    mock_builder: MockBuilder,
) -> Callable[[str], Awaitable[None]]:
    async def start(bake_id: str) -> None:
        async with BatchExecutor.create(
            get_console(),
            bake_id,
            patched_client,
            batch_storage,
            polling_timeout=0.05,
            run_builder_job=mock_builder.run,
        ) as executor:
            await executor.run()

    return start


RunExecutor = Callable[
    [Path, str], Awaitable[Tuple[Dict[Tuple[str, ...], Task], TaskStatus]]
]


@pytest.fixture()
def run_executor(
    make_batch_runner: MakeBatchRunner,
    start_locals_executor: Callable[[str], Awaitable[None]],
    start_executor: Callable[[str], Awaitable[None]],
    batch_storage: Storage,
) -> RunExecutor:
    async def run(
        config_loc: Path, batch_name: str, args: Optional[Mapping[str, str]] = None
    ) -> Tuple[Dict[Tuple[str, ...], Task], TaskStatus]:
        runner = await make_batch_runner(config_loc)
        bake, _ = await runner._setup_bake(batch_name, args)
        status = await start_locals_executor(bake.id)
        if status == TaskStatus.SUCCEEDED:
            await start_executor(bake.id)
        attempt_storage = batch_storage.bake(id=bake.id).last_attempt()
        attempt = await attempt_storage.get()
        tasks = {task.yaml_id: task async for task in attempt_storage.list_tasks()}
        return tasks, attempt.result

    return run


async def test_simple_batch_ok(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-seq"))
    task_descr = await jobs_mock.get_task("task-1")
    assert task_descr.container.image.name == "ubuntu"
    assert task_descr.container.command
    assert "echo abc" in task_descr.container.command
    await jobs_mock.mark_done("task-1")

    task_descr = await jobs_mock.get_task("task-2")
    assert task_descr.container.image.name == "ubuntu"
    assert task_descr.container.command
    assert "echo def" in task_descr.container.command
    await jobs_mock.mark_done("task-2")
    await executor_task


async def test_complex_seq(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-complex-seq"))
    await jobs_mock.get_task("task_1_a")
    await jobs_mock.get_task("task_1_b")

    await jobs_mock.mark_started("task_1_a")
    await jobs_mock.mark_started("task_1_b")

    await jobs_mock.get_task("task_2_at_running.task_1")
    await jobs_mock.mark_done(
        "task_2_at_running.task_1", b"::set-output name=task1::Task 1 val 1"
    )
    await jobs_mock.get_task("task_2_at_running.task_2")

    await jobs_mock.mark_done("task_1_a")
    await jobs_mock.mark_done("task_1_b")

    await jobs_mock.get_task("task_2_after")
    await jobs_mock.mark_done("task_2_after")

    with pytest.raises(AssertionError, match=r".* did not appeared in timeout of .*"):
        await jobs_mock.get_task("task_3", timeout=0.5)

    await jobs_mock.mark_started("task_2_at_running.task_2")

    await jobs_mock.get_task("task_3")
    await jobs_mock.mark_done("task_3")

    await jobs_mock.mark_done(
        "task_2_at_running.task_2", b"::set-output name=task2::Task 2 val 2"
    )

    await asyncio.wait_for(executor_task, timeout=5)


class FakeForceStopException(BaseException):
    # Used to simulate almost instant kills of executor
    pass


def run_in_loop(coro: Awaitable[Any]) -> Any:
    loop = asyncio.new_event_loop()
    return loop.run_until_complete(coro)


async def test_complex_seq_continue(
    jobs_mock: JobsMock,
    start_executor: Callable[[str], Awaitable[None]],
    batch_runner: BatchRunner,
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-complex-seq")
    # Use separate thread to allow force abort
    executor_task = asyncio.get_event_loop().run_in_executor(
        None, run_in_loop, start_executor(bake.id)
    )

    await jobs_mock.get_task("task_1_a")
    await jobs_mock.get_task("task_1_b")

    await jobs_mock.mark_started("task_1_a")
    await jobs_mock.mark_started("task_1_b")

    await jobs_mock.get_task("task_2_at_running.task_1")
    await jobs_mock.mark_done(
        "task_2_at_running.task_1", b"::set-output name=task1::Task 1 val 1"
    )
    await jobs_mock.get_task("task_2_at_running.task_2")

    await jobs_mock.mark_done("task_1_a")
    await jobs_mock.mark_done("task_1_b")

    await jobs_mock.get_task("task_2_after")
    await jobs_mock.mark_done("task_2_after")

    jobs_mock.exception_on_status = FakeForceStopException()
    try:
        await executor_task
    except FakeForceStopException:
        pass
    jobs_mock.exception_on_status = None

    executor_task_2 = asyncio.ensure_future(start_executor(bake.id))

    await jobs_mock.mark_started("task_2_at_running.task_2")

    await jobs_mock.get_task("task_3")
    await jobs_mock.mark_done("task_3")

    await jobs_mock.mark_done(
        "task_2_at_running.task_2", b"::set-output name=task2::Task 2 val 2"
    )

    await asyncio.wait_for(executor_task_2, timeout=5)


async def test_complex_seq_continue_2(
    jobs_mock: JobsMock,
    start_executor: Callable[[str], Awaitable[None]],
    batch_runner: BatchRunner,
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-complex-seq")
    # Use separate thread to allow force abort
    executor_task = asyncio.get_event_loop().run_in_executor(
        None, run_in_loop, start_executor(bake.id)
    )

    await jobs_mock.get_task("task_1_a")
    await jobs_mock.get_task("task_1_b")

    await jobs_mock.mark_started("task_1_a")
    await jobs_mock.mark_started("task_1_b")

    await jobs_mock.get_task("task_2_at_running.task_1")
    await jobs_mock.mark_done(
        "task_2_at_running.task_1", b"::set-output name=task1::Task 1 val 1"
    )
    await jobs_mock.get_task("task_2_at_running.task_2")
    await jobs_mock.mark_started("task_2_at_running.task_2")

    jobs_mock.exception_on_status = FakeForceStopException()
    try:
        await executor_task
    except FakeForceStopException:
        pass
    jobs_mock.exception_on_status = None

    await jobs_mock.mark_done(
        "task_2_at_running.task_2", b"::set-output name=task2::Task 2 val 2"
    )

    executor_task_2 = asyncio.ensure_future(start_executor(bake.id))

    await jobs_mock.mark_done("task_1_a")
    await jobs_mock.mark_done("task_1_b")

    await jobs_mock.get_task("task_2_after")
    await jobs_mock.mark_done("task_2_after")

    await jobs_mock.get_task("task_3")
    await jobs_mock.mark_done("task_3")

    await asyncio.wait_for(executor_task_2, timeout=5)


async def test_complex_seq_restart(
    jobs_mock: JobsMock,
    start_executor: Callable[[str], Awaitable[None]],
    batch_runner: BatchRunner,
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-complex-seq")
    executor_task = asyncio.ensure_future(start_executor(bake.id))

    await jobs_mock.get_task("task_1_a")
    await jobs_mock.get_task("task_1_b")

    await jobs_mock.mark_started("task_1_a")
    await jobs_mock.mark_started("task_1_b")

    await jobs_mock.get_task("task_2_at_running.task_1")
    await jobs_mock.mark_done(
        "task_2_at_running.task_1", b"::set-output name=task1::Task 1 val 1"
    )
    await jobs_mock.get_task("task_2_at_running.task_2")

    await jobs_mock.mark_done("task_1_a")
    await jobs_mock.mark_done("task_1_b")

    await jobs_mock.get_task("task_2_after")
    await jobs_mock.mark_done("task_2_after")

    await asyncio.sleep(0.1)  # Let loop proceed

    await jobs_mock.mark_failed("task_2_at_running.task_2")

    await asyncio.wait_for(executor_task, timeout=5)

    jobs_mock.clear()
    data, _ = await batch_runner._restart(bake.id)

    executor_task = asyncio.ensure_future(start_executor(bake.id))

    await jobs_mock.get_task("task_2_at_running.task_2")
    await jobs_mock.mark_started("task_2_at_running.task_2")

    await jobs_mock.get_task("task_3")
    await jobs_mock.mark_done("task_3")

    await jobs_mock.mark_done(
        "task_2_at_running.task_2", b"::set-output name=task2::Task 2 val 2"
    )

    await asyncio.wait_for(executor_task, timeout=5)


async def test_workdir_passed(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-workdir"))
    task_descr = await jobs_mock.get_task("task-1")
    assert task_descr.container.working_dir == "/test/workdir"
    await jobs_mock.mark_done("task-1")

    await executor_task


async def test_batch_with_action_ok(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-action-call"))
    await jobs_mock.get_task("test.task-1")
    await jobs_mock.mark_done("test.task-1", b"::set-output name=task1::Task 1 val 1")

    await jobs_mock.get_task("test.task-2")
    await jobs_mock.mark_done("test.task-2", b"::set-output name=task2::Task 2 value 2")
    await executor_task


async def test_batch_with_module_ok(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "batch_module", "batch-module-call")
    )
    task = await jobs_mock.get_task("test.task-1")
    assert task.container.command
    assert "batch_module_call" in task.container.command
    assert task.container.env.get("TEST") == "test_value"
    assert "test-tag" in task.tags
    assert task.preset_name == "test-preset"
    assert task.life_span == timedelta(days=2).total_seconds()
    assert task.schedule_timeout == timedelta(minutes=60).total_seconds()
    assert task.container.working_dir == "/some/dir"
    assert task.container.volumes[0].container_path == "/volume"
    await jobs_mock.mark_done("test.task-1", b"::set-output name=task1::Task 1 val 1")

    task = await jobs_mock.get_task("test.task-2")
    assert task.container.command
    assert "batch_module_call" in task.container.command
    assert task.container.env.get("TEST") == "test_value"
    await jobs_mock.mark_done("test.task-2", b"::set-output name=task2::Task 2 value 2")
    await executor_task


@pytest.mark.parametrize(
    "batch_name,vars",
    [
        ("batch-matrix-1", [("o1", "t1"), ("o1", "t2"), ("o2", "t1"), ("o2", "t2")]),
        ("batch-matrix-2", [("o1", "t2"), ("o2", "t1"), ("o2", "t2")]),
        (
            "batch-matrix-3",
            [("o1", "t1"), ("o1", "t2"), ("o2", "t1"), ("o2", "t2"), ("o3", "t3")],
        ),
        ("batch-matrix-4", [("o2", "t1"), ("o2", "t2"), ("o3", "t3")]),
        ("batch-matrix-5", [("o2", "t1"), ("o2", "t2"), ("o1", "t1")]),
    ],
)
async def test_batch_matrix(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
    batch_name: str,
    vars: List[Tuple[str, str]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, batch_name))
    for var_1, var_2 in vars:
        descr = await jobs_mock.get_task(f"task-1-{var_1}-{var_2}")
        assert descr.container.command
        assert f"{var_1}-{var_2}" in descr.container.command
        await jobs_mock.mark_done(f"task-1-{var_1}-{var_2}")

    await executor_task


@pytest.mark.parametrize(
    "batch_name,max_parallel,vars",
    [
        pytest.param(
            "batch-matrix-max-parallel",
            2,
            [("o1", "t1"), ("o1", "t2"), ("o2", "t1"), ("o2", "t2")],
            marks=pytest.mark.xfail(reason="Local max_parallel is broken"),
        ),
        (
            "batch-matrix-max-parallel-global",
            2,
            [("o1", "t1"), ("o1", "t2"), ("o2", "t1"), ("o2", "t2")],
        ),
    ],
)
async def test_batch_matrix_max_parallel(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
    batch_name: str,
    max_parallel: int,
    vars: List[Tuple[str, str]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, batch_name))
    done, pending = await asyncio.wait(
        [jobs_mock.get_task(f"task-1-{var_1}-{var_2}") for var_1, var_2 in vars],
        return_when=asyncio.ALL_COMPLETED,
    )

    alive_cnt = sum(1 for task in done if task.exception() is None)
    assert alive_cnt == max_parallel
    executor_task.cancel()


async def test_enable_cancels_depending_tasks(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    await asyncio.wait_for(run_executor(assets, "batch-first-disabled"), timeout=5)


async def test_disabled_task_is_not_required(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets, "batch-disabled-not-needed")
    )

    await jobs_mock.mark_done("task-1")
    await jobs_mock.mark_done("task-3")
    await jobs_mock.mark_done("task-4")

    await executor_task


async def test_cancellation(
    jobs_mock: JobsMock,
    batch_runner: BatchRunner,
    start_executor: Callable[[str], Awaitable[None]],
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-seq")
    executor_task = asyncio.ensure_future(start_executor(bake.id))
    await jobs_mock.mark_done("task-1")
    descr = await jobs_mock.get_task("task-2")
    assert descr.status == JobStatus.PENDING

    await batch_runner.cancel(bake.id)
    await executor_task

    descr = await jobs_mock.get_task("task-1")
    assert descr.status == JobStatus.SUCCEEDED

    descr = await jobs_mock.get_task("task-2")
    assert descr.status == JobStatus.CANCELLED


async def test_volumes_parsing(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-volumes-parsing"))

    descr = await jobs_mock.get_task("task-1")

    storage_volume = descr.container.volumes[0]
    assert "folder" in storage_volume.storage_uri.path
    assert storage_volume.container_path == "/mnt/storage"
    assert not storage_volume.read_only

    disk_volume = descr.container.disk_volumes[0]
    assert "disk-name" in disk_volume.disk_uri.path
    assert disk_volume.container_path == "/mnt/disk"
    assert disk_volume.read_only

    secret_file = descr.container.secret_files[0]
    assert "key" in secret_file.secret_uri.path
    assert secret_file.container_path == "/mnt/secret"

    await jobs_mock.mark_done("task-1")
    await executor_task


async def test_always_after_disabled(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets, "batch-disabled-but-always")
    )

    await jobs_mock.mark_done("task-2")

    await executor_task


async def test_always_during_cancellation(
    jobs_mock: JobsMock,
    batch_runner: BatchRunner,
    start_executor: Callable[[str], Awaitable[None]],
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-last-always")
    executor_task = asyncio.ensure_future(start_executor(bake.id))
    descr = await jobs_mock.get_task("task-1")
    assert descr.status == JobStatus.PENDING

    await batch_runner.cancel(bake.id)

    await jobs_mock.mark_done("task-3")

    await executor_task


@pytest.mark.skipif(
    sys.platform == "win32", reason="cp command is not support by Windows"
)
async def test_local_action(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    with TemporaryDirectory() as dir:
        ws = LocalPath(dir) / "local_actions"
        shutil.copytree(assets / "local_actions", ws)

        executor_task = asyncio.ensure_future(run_executor(ws, "call-cp"))
        descr = await jobs_mock.get_task("remote-task")
        assert descr.container.command
        assert "echo 0" in descr.container.command

        assert (ws / "file").read_text() == "test\n"
        assert (ws / "file_copy").read_text() == "test\n"

        await jobs_mock.mark_done("remote-task")

        await executor_task


@pytest.mark.skipif(
    sys.platform == "win32", reason="cp command is not support by Windows"
)
async def test_action_with_local_action(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    with TemporaryDirectory() as dir:
        ws = LocalPath(dir) / "local_actions"
        shutil.copytree(assets / "local_actions", ws)

        executor_task = asyncio.ensure_future(
            run_executor(ws, "call-batch-action-with-local")
        )
        descr = await jobs_mock.get_task("call_action.remote-task")
        assert descr.container.command
        assert "echo 0" in descr.container.command

        assert (ws / "file").read_text() == "test\n"
        assert (ws / "file_copy").read_text() == "test\n"

        await jobs_mock.mark_done("call_action.remote-task")

        await executor_task


async def test_stateful_no_post(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "stateful_actions", "call-no-post")
    )

    await jobs_mock.mark_done("first")
    await jobs_mock.mark_done("test")
    await jobs_mock.mark_done("last")

    await executor_task


async def test_stateful_with_post(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "stateful_actions", "call-with-post")
    )

    await jobs_mock.mark_done("first")
    await jobs_mock.mark_done("test")
    await jobs_mock.mark_done("last")
    await jobs_mock.mark_done("post-test")

    await executor_task


async def test_stateful_with_state(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "stateful_actions", "call-with-state")
    )

    await jobs_mock.mark_done("first")
    await jobs_mock.mark_done("test", b"::save-state name=const::constant")
    await jobs_mock.mark_done("last")
    await jobs_mock.mark_done("post-test")

    await executor_task


async def test_stateful_post_after_fail(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "stateful_actions", "call-with-post")
    )

    await jobs_mock.mark_done("first")
    await jobs_mock.mark_done("test")
    await jobs_mock.mark_failed("last")
    await jobs_mock.mark_done("post-test")

    await executor_task


async def test_stateful_post_after_cancellation(
    jobs_mock: JobsMock,
    assets: Path,
    make_batch_runner: MakeBatchRunner,
    start_executor: Callable[[str], Awaitable[None]],
) -> None:
    runner = await make_batch_runner(assets / "stateful_actions")
    bake, _ = await runner._setup_bake("call-with-post")
    executor_task = asyncio.ensure_future(start_executor(bake.id))
    await jobs_mock.mark_done("first")
    await jobs_mock.mark_done("test")

    await runner.cancel(bake.id)

    await jobs_mock.mark_done("post-test")

    await executor_task


async def test_restart(batch_storage: Storage, batch_runner: BatchRunner) -> None:
    bake, _ = await batch_runner._setup_bake("batch-seq")

    attempt_storage = batch_storage.bake(id=bake.id).last_attempt()
    job_1 = ("task-1",)
    job_2 = ("task-2",)
    task1 = await attempt_storage.create_task(
        yaml_id=job_1,
        raw_id="test-1",
        status=TaskStatus.SUCCEEDED,
        outputs={"key": "value"},
        state={},
    )
    await attempt_storage.create_task(
        yaml_id=job_2,
        raw_id="test-2",
        status=TaskStatus.FAILED,
        outputs={},
        state={},
    )
    await attempt_storage.update(result=TaskStatus.FAILED)

    bake_restart, _ = await batch_runner._restart(bake.id)
    assert bake_restart.id == bake.id

    attempt_storage = batch_storage.bake(id=bake.id).last_attempt()
    attempt2 = await attempt_storage.get()
    assert attempt2.number == 2
    tasks_map = {task.yaml_id: task async for task in attempt_storage.list_tasks()}
    assert job_1 in tasks_map
    assert job_2 not in tasks_map
    task1_new = tasks_map[job_1]
    assert task1_new.outputs == task1.outputs


async def test_restart_not_last_attempt(
    batch_storage: Storage, batch_runner: BatchRunner
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-seq")
    attempt_storage = batch_storage.bake(id=bake.id).last_attempt()
    job_1 = ("task-1",)
    job_2 = ("task-2",)
    task1 = await attempt_storage.create_task(
        yaml_id=job_1,
        raw_id="test-1",
        status=TaskStatus.SUCCEEDED,
        outputs={"key": "value"},
        state={},
    )
    await attempt_storage.create_task(
        yaml_id=job_2,
        raw_id="test-2",
        status=TaskStatus.FAILED,
        outputs={},
        state={},
    )
    await attempt_storage.update(result=TaskStatus.FAILED)

    await batch_runner._restart(bake.id)
    with pytest.raises(Exception):
        await batch_runner._restart(bake.id, attempt_no=1)

    await batch_storage.bake(id=bake.id).last_attempt().update(result=TaskStatus.FAILED)

    bake_restart, _ = await batch_runner._restart(bake.id, attempt_no=1)
    assert bake_restart.id == bake.id

    attempt_storage = batch_storage.bake(id=bake.id).last_attempt()
    attempt3 = await attempt_storage.get()
    assert attempt3.number == 3
    tasks_map = {task.yaml_id: task async for task in attempt_storage.list_tasks()}
    assert job_1 in tasks_map
    assert job_2 not in tasks_map
    task1_new = tasks_map[job_1]
    assert task1_new.outputs == task1.outputs


async def test_fully_cached_simple(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: RunExecutor,
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-seq"))

    await jobs_mock.mark_done("task-1")
    await jobs_mock.mark_done("task-2")
    await executor_task

    tasks, status = await asyncio.wait_for(
        run_executor(assets, "batch-seq"), timeout=10
    )
    assert status == TaskStatus.SUCCEEDED
    for task in tasks.values():
        assert task.status == TaskStatus.CACHED


async def test_fully_cached_with_action(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: RunExecutor,
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-action-call"))
    await jobs_mock.mark_done("test.task-1", b"::set-output name=task1::Task 1 val 1")
    await jobs_mock.mark_done("test.task-2", b"::set-output name=task2::Task 2 value 2")
    await executor_task

    finished, status = await asyncio.wait_for(
        run_executor(assets, "batch-action-call"), timeout=10
    )

    assert status == TaskStatus.SUCCEEDED
    assert finished[("test", "task_1")].status == TaskStatus.CACHED
    assert finished[("test", "task_2")].status == TaskStatus.CACHED


async def test_cached_same_needs(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: RunExecutor,
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-test-cache"))

    await jobs_mock.mark_done("task-1", b"::set-output name=arg::val")
    await jobs_mock.mark_done("task-2")
    await executor_task

    jobs_mock.clear()
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-test-cache"))

    await jobs_mock.mark_done("task-1", b"::set-output name=arg::val")

    finished, status = await asyncio.wait_for(executor_task, timeout=10)

    assert status == TaskStatus.SUCCEEDED
    assert finished[("task-1",)].status == TaskStatus.SUCCEEDED
    assert finished[("task-2",)].status == TaskStatus.CACHED


async def test_not_cached_if_different_needs(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-test-cache"))

    await jobs_mock.mark_done("task-1", b"::set-output name=arg::val")
    await jobs_mock.mark_done("task-2")
    await executor_task

    jobs_mock.clear()
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-test-cache"))

    await jobs_mock.mark_done("task-1", b"::set-output name=arg::val2")
    await jobs_mock.mark_done("task-2")
    await executor_task


async def test_batch_params(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str, Mapping[str, str]], Awaitable[None]],
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets, "batch-params-required", {"arg2": "test_value"})
    )
    task_descr = await jobs_mock.get_task("task-1")
    assert set(task_descr.tags).issuperset(
        {
            "flow:batch-params-required",
            "test_value",
            "val1",
            "project:unit",
            "task:task-1",
        }
    )
    await jobs_mock.mark_done("task-1")

    await executor_task


async def test_batch_bake_id_tag(
    jobs_mock: JobsMock,
    start_executor: Callable[[str], Awaitable[None]],
    batch_runner: BatchRunner,
) -> None:
    bake, _ = await batch_runner._setup_bake("batch-seq")
    executor_task = asyncio.ensure_future(start_executor(bake.id))

    task_descr = await jobs_mock.get_task("task-1")
    assert set(task_descr.tags) == {
        "flow:batch-seq",
        f"bake_id:{bake.id}",
        "project:unit",
        "task:task-1",
    }
    await jobs_mock.mark_done("task-1")

    await jobs_mock.mark_done("task-2")

    await executor_task


async def test_bake_marked_as_failed_on_eval_error(
    assets: Path,
    run_executor: RunExecutor,
) -> None:
    _, status = await run_executor(assets, "batch-bad-eval")
    assert status == TaskStatus.FAILED


async def test_bake_marked_as_failed_on_eval_error_in_local(
    assets: Path,
    run_executor: RunExecutor,
) -> None:
    _, status = await run_executor(assets / "local_actions", "call-bad-local")
    assert status == TaskStatus.FAILED


async def test_bake_marked_as_failed_on_random_error(
    assets: Path,
    run_executor: RunExecutor,
    jobs_mock: JobsMock,
) -> None:
    jobs_mock.exception_on_status = Exception("Something failed")
    _, status = await run_executor(assets, "batch-seq")
    assert status == TaskStatus.FAILED


async def test_bake_marked_as_cancelled_on_task_cancelation(
    assets: Path,
    run_executor: RunExecutor,
) -> None:
    executor_task = asyncio.ensure_future(run_executor(assets, "batch-seq"))
    await asyncio.sleep(0)  # All to start
    executor_task.cancel()
    _, status = await executor_task
    assert status == TaskStatus.CANCELLED


async def _wait_for_build(mock_builder: MockBuilder, ref: str) -> str:
    async def _waiter() -> None:
        while ref not in mock_builder.ref2job:
            await asyncio.sleep(0.1)

    await asyncio.wait_for(_waiter(), timeout=1)
    return mock_builder.ref2job[ref]


async def test_image_builds(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
    mock_builder: MockBuilder,
    batch_storage: Storage,
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "batch_images", "batch")
    )

    job_id = await _wait_for_build(mock_builder, "image:banana1")

    project_storage = batch_storage.project(yaml_id="batch_images")
    bake = [bake async for bake in project_storage.list_bakes()][0]
    bake_storage = project_storage.bake(id=bake.id)
    image = await bake_storage.bake_image(ref="image:banana1").get()
    assert mock_builder.runs[-1] == (
        "neuro-extras",
        "image",
        "build",
        "--file=Dockerfile",
        str(image.context_on_storage),
        "image:banana1",
    )
    await jobs_mock.mark_done(job_id)

    task = await jobs_mock.get_task("action.task_1")
    assert task.container.image.name == "banana1"
    await jobs_mock.mark_done("action.task_1")

    job_id = await _wait_for_build(mock_builder, "image:banana2")
    assert mock_builder.runs[-1] == (
        "neuro-extras",
        "image",
        "build",
        "--file=Dockerfile",
        "storage://default/foo/val1",
        "image:banana2",
    )
    await jobs_mock.mark_done(job_id)

    task = await jobs_mock.get_task("action.task_2")
    assert task.container.image.name == "banana2"
    await jobs_mock.mark_done("action.task_2")

    job_id = await _wait_for_build(mock_builder, "image:main")
    image = await bake_storage.bake_image(ref="image:main").get()
    assert mock_builder.runs[-1] == (
        "neuro-extras",
        "image",
        "build",
        "--file=Dockerfile",
        str(image.context_on_storage),
        "image:main",
    )
    await jobs_mock.mark_done(job_id)

    task = await jobs_mock.get_task("task")
    assert task.container.image.name == "main"
    await jobs_mock.mark_done("task")

    await asyncio.wait_for(executor_task, timeout=1)


async def test_image_builds_skip_if_present(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
    images_mock: ImagesMock,
    client: Client,
    batch_storage: Storage,
) -> None:
    images_mock.known_images = {
        f"image://{client.cluster_name}/{client.username}/banana1:latest": Tag(
            "latest"
        ),
        f"image://{client.cluster_name}/{client.username}/banana2:latest": Tag(
            "latest"
        ),
        f"image://{client.cluster_name}/{client.username}/main:latest": Tag("latest"),
    }

    executor_task = asyncio.ensure_future(
        run_executor(assets / "batch_images", "batch")
    )

    await jobs_mock.get_task("action.task_1")
    await jobs_mock.mark_done("action.task_1")

    await jobs_mock.get_task("action.task_2")
    await jobs_mock.mark_done("action.task_2")

    await jobs_mock.get_task("task")
    await jobs_mock.mark_done("task")

    await asyncio.wait_for(executor_task, timeout=1)

    project_storage = batch_storage.project(yaml_id="batch_images")
    bake = [bake async for bake in project_storage.list_bakes()][0]
    bake_storage = project_storage.bake(id=bake.id)
    image = await bake_storage.bake_image(ref="image:main").get()
    assert image.status == ImageStatus.CACHED


async def test_image_builds_if_present_but_force(
    jobs_mock: JobsMock,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
    images_mock: ImagesMock,
    mock_builder: MockBuilder,
    batch_storage: Storage,
    client: Client,
) -> None:
    images_mock.known_images = {
        f"image://{client.cluster_name}/{client.username}/banana1:latest": Tag(
            "latest"
        ),
        f"image://{client.cluster_name}/{client.username}/banana2:latest": Tag(
            "latest"
        ),
        f"image://{client.cluster_name}/{client.username}/main:latest": Tag("latest"),
    }

    executor_task = asyncio.ensure_future(
        run_executor(assets / "batch_images", "batch-with-force")
    )

    await jobs_mock.get_task("action.task_1")
    await jobs_mock.mark_done("action.task_1")

    await jobs_mock.get_task("action.task_2")
    await jobs_mock.mark_done("action.task_2")

    job_id = await _wait_for_build(mock_builder, "image:main")
    project_storage = batch_storage.project(yaml_id="batch_images")
    bake = [bake async for bake in project_storage.list_bakes()][0]
    bake_storage = project_storage.bake(id=bake.id)
    image = await bake_storage.bake_image(ref="image:main").get()
    assert mock_builder.runs[-1] == (
        "neuro-extras",
        "image",
        "build",
        "--file=Dockerfile",
        "--force-overwrite",
        str(image.context_on_storage),
        "image:main",
    )
    await jobs_mock.mark_done(job_id)

    await jobs_mock.get_task("task")
    await jobs_mock.mark_done("task")

    await asyncio.wait_for(executor_task, timeout=1)


async def test_image_builds_cancel(
    jobs_mock: JobsMock,
    make_batch_runner: MakeBatchRunner,
    assets: Path,
    run_executor: Callable[[Path, str], Awaitable[None]],
    mock_builder: MockBuilder,
    batch_storage: Storage,
) -> None:
    executor_task = asyncio.ensure_future(
        run_executor(assets / "batch_images", "batch")
    )
    batch_runner = await make_batch_runner(assets / "batch_images")

    async def _wait_for_build(ref: str) -> str:
        async def _waiter() -> None:
            while ref not in mock_builder.ref2job:
                await asyncio.sleep(0.1)

        await asyncio.wait_for(_waiter(), timeout=1)
        return mock_builder.ref2job[ref]

    job_id = await _wait_for_build("image:banana1")
    project_storage = batch_storage.project(yaml_id="batch_images")
    bake = [bake async for bake in project_storage.list_bakes()][0]
    await batch_runner.cancel(bake.id)

    await asyncio.wait_for(executor_task, timeout=1)

    descr = await jobs_mock.get_task(job_id)
    assert descr.status == JobStatus.CANCELLED
